import json
from typing import Dict

import pandas as pd
import numpy as np
import torch
from pathlib import Path
from itertools import repeat
from collections import OrderedDict
from collections.abc import Mapping
from scvelo.core import l2_norm, prod_sum, sum


def ensure_dir(dirname):
    dirname = Path(dirname)
    if not dirname.is_dir():
        dirname.mkdir(parents=True, exist_ok=False)
    return dirname


def read_json(fname):
    fname = Path(fname)
    with fname.open("rt") as handle:
        return json.load(handle, object_hook=OrderedDict)


def write_json(content, fname):
    fname = Path(fname)
    with fname.open("wt") as handle:
        json.dump(content, handle, indent=4, sort_keys=False)


def save_model_and_config(model, config, directory):
    """saves model and config to directory."""
    directory = ensure_dir(directory)
    torch.save(model.state_dict(), directory / "model.pt")
    write_json(config, directory / "config.json")


def inf_loop(data_loader):
    """wrapper function for endless data loader."""
    for loader in repeat(data_loader):
        yield from loader


def validate_config(config: Mapping) -> Mapping:
    """
    Return config if it is valid, otherwise raise an error.
    """

    # check if the gpu verion of dgl is installed
    if config["n_gpu"] > 0:
        import dgl

        try:
            dgl.graph([]).to("cuda")
        except dgl.DGLError:
            print(
                "Config Warning: Set to use GPU, but GPU version of DGL is not "
                "installed. Reset to use CPU instead."
            )
            config["n_gpu"] = 0

    return config


def update_dict(d: Dict, u: Mapping, copy=False):
    """recursively updates nested dict with values from u."""
    if copy:
        d = d.copy()
    for k, v in u.items():
        if isinstance(v, Mapping):
            r = update_dict(d.get(k, {}), v, copy)
            d[k] = r
        else:
            d[k] = u[k]
    return d


def get_indices(dist, n_neighbors=None, mode_neighbors="distances"):
    from scvelo.preprocessing.neighbors import compute_connectivities_umap

    D = dist.copy()
    D.data += 1e-6

    n_counts = sum(D > 0, axis=1)
    n_neighbors = (
        n_counts.min() if n_neighbors is None else min(n_counts.min(), n_neighbors)
    )
    rows = np.where(n_counts > n_neighbors)[0]
    cumsum_neighs = np.insert(n_counts.cumsum(), 0, 0)
    dat = D.data

    for row in rows:
        n0, n1 = cumsum_neighs[row], cumsum_neighs[row + 1]
        rm_idx = n0 + dat[n0:n1].argsort()[n_neighbors:]
        dat[rm_idx] = 0
    D.eliminate_zeros()

    D.data -= 1e-6
    if mode_neighbors == "distances":
        indices = D.indices.reshape((-1, n_neighbors))
    elif mode_neighbors == "connectivities":
        knn_indices = D.indices.reshape((-1, n_neighbors))
        knn_distances = D.data.reshape((-1, n_neighbors))
        _, conn = compute_connectivities_umap(
            knn_indices, knn_distances, D.shape[0], n_neighbors
        )
        indices = get_indices_from_csr(conn)
    return indices, D


def get_indices_from_csr(conn):
    # extracts indices from connectivity matrix, pads with nans
    ixs = np.ones((conn.shape[0], np.max((conn > 0).sum(1)))) * np.nan
    for i in range(ixs.shape[0]):
        cell_indices = conn[i, :].indices
        ixs[i, : len(cell_indices)] = cell_indices
    return ixs


class MetricTracker:
    def __init__(self, *keys, writer=None):
        self.writer = writer
        self._data = pd.DataFrame(index=keys, columns=["total", "counts", "average"])
        self.reset()

    def reset(self):
        for col in self._data.columns:
            self._data[col].values[:] = 0

    def update(self, key, value, n=1):
        if self.writer is not None:
            self.writer.add_scalar(key, value)
        self._data.total[key] += value * n
        self._data.counts[key] += n
        self._data.average[key] = self._data.total[key] / self._data.counts[key]

    def avg(self, key):
        return self._data.average[key]

    def result(self):
        return dict(self._data.average)



import torch


def accuracy(output, target):
    with torch.no_grad():
        pred = torch.argmax(output, dim=1)
        assert pred.shape[0] == len(target)
        correct = 0
        correct += torch.sum(pred == target).item()
    return correct / len(target)


def top_k_acc(output, target, k=3):
    with torch.no_grad():
        pred = torch.topk(output, k, dim=1)[1]
        assert pred.shape[0] == len(target)
        correct = 0
        for i in range(k):
            correct += torch.sum(pred[:, i] == target).item()
    return correct / len(target)


def mse(output, target):
    with torch.no_grad():
        se = torch.mean((target - output) ** 2)
    return se


def min_mse(output, target):
    """
    selects one closest cell and computes the loss

    the target is the set of velocity target candidates,
    find the closest in them.

    output: torch.tensor e.g. (128, 2000)
    target: torch.tensor e.g. (128, 30, 2000)
    """
    with torch.no_grad():
        distance = torch.pow(
            target - torch.unsqueeze(output, 1), exponent=2
        )  # (128, 30, 2000)
        distance = torch.sum(distance, dim=2)  # (128, 30)
        min_distance = torch.min(distance, dim=1)[0]  # (128,)

        # loss = torch.mean(torch.max(torch.tensor(alpha).float(), min_distance))
        se = torch.mean(min_distance)

    return se



import logging
import logging.config
from pathlib import Path
from deepvelo.utils import read_json


def setup_logging(
    save_dir, log_config="logger/logger_config.json", default_level=logging.INFO
):
    """
    Setup logging configuration
    """
    log_config = Path(log_config)
    if log_config.is_file():
        config = read_json(log_config)
        # modify logging paths based on run config
        for _, handler in config["handlers"].items():
            if "filename" in handler:
                handler["filename"] = str(save_dir / handler["filename"])

        logging.config.dictConfig(config)
    else:
        print(
            "Warning: logging configuration file is not found in {}.".format(log_config)
        )
        logging.basicConfig(level=default_level)



import os
import logging
from pathlib import Path
from functools import reduce, partial
from operator import getitem
from datetime import datetime
from typing import Callable
from deepvelo.logger import setup_logging
from deepvelo.utils import read_json, write_json, validate_config


class ConfigParser:
    def __init__(
        self,
        config,
        resume=None,
        modification=None,
        run_id=None,
        validator: Callable = validate_config,
    ):
        """
        class to parse configuration json file. Handles hyperparameters for training, initializations of modules, checkpoint saving
        and logging module.
        :param config: Dict containing configurations, hyperparameters for training. contents of `config.json` file for example.
        :param resume: String, path to the checkpoint being loaded.
        :param modification: Dict keychain:value, specifying position values to be replaced from config dict.
        :param run_id: Unique Identifier for training processes. Used to save checkpoints and training log. Timestamp is being used as default
        """
        # load config file and apply modification
        self._config = _update_config(config, modification)
        self.resume = resume

        if validator:
            self._config = validator(self._config)

        # set save_dir where trained model and log will be saved.
        save_dir = Path(self.config["trainer"]["save_dir"])

        exper_name = self.config["name"]
        if run_id is None:  # use timestamp as default run-id
            run_id = datetime.now().strftime(r"%m%d_%H%M%S")
        self._save_dir = save_dir / "models" / exper_name / run_id
        self._log_dir = save_dir / "log" / exper_name / run_id

        # make directory for saving checkpoints and log.
        exist_ok = run_id == ""
        self.save_dir.mkdir(parents=True, exist_ok=exist_ok)
        self.log_dir.mkdir(parents=True, exist_ok=exist_ok)

        # save updated config file to the checkpoint dir
        write_json(self.config, self.save_dir / "config.json")

        # configure logging module
        setup_logging(self.log_dir)
        self.log_levels = {0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}

    @classmethod
    def from_args(cls, args, options=""):
        """
        Initialize this class from some cli arguments. Used in train, test.
        """
        for opt in options:
            args.add_argument(*opt.flags, default=None, type=opt.type)
        if not isinstance(args, tuple):
            args = args.parse_args()

        if args.device is not None:
            os.environ["CUDA_VISIBLE_DEVICES"] = args.device
        if args.resume is not None:
            resume = Path(args.resume)
            cfg_fname = resume.parent / "config.json"
        else:
            msg_no_cfg = "Configuration file need to be specified. Add '-c config.json', for example."
            assert args.config is not None, msg_no_cfg
            resume = None
            cfg_fname = Path(args.config)

        config = read_json(cfg_fname)
        if args.config and resume:
            # update new config for fine-tuning
            config.update(read_json(args.config))

        # parse custom cli options into dictionary
        modification = {
            opt.target: getattr(args, _get_opt_name(opt.flags)) for opt in options
        }
        return cls(config, resume, modification)

    def init_obj(self, name, module, *args, **kwargs):
        """
        Finds a function handle with the name given as 'type' in config, and returns the
        instance initialized with corresponding arguments given.

        `object = config.init_obj('name', module, a, b=1)`
        is equivalent to
        `object = module.name(a, b=1)`
        """
        module_name = self[name]["type"]
        module_args = dict(self[name]["args"])
        assert all(
            [k not in module_args for k in kwargs]
        ), "Overwriting kwargs given in config file is not allowed"
        module_args.update(kwargs)
        return getattr(module, module_name)(*args, **module_args)

    def init_ftn(self, name, module, *args, **kwargs):
        """
        Finds a function handle with the name given as 'type' in config, and returns the
        function with given arguments fixed with functools.partial.

        `function = config.init_ftn('name', module, a, b=1)`
        is equivalent to
        `function = lambda *args, **kwargs: module.name(a, *args, b=1, **kwargs)`.
        """
        module_name = self[name]["type"]
        module_args = dict(self[name]["args"])
        assert all(
            [k not in module_args for k in kwargs]
        ), "Overwriting kwargs given in config file is not allowed"
        module_args.update(kwargs)
        return partial(getattr(module, module_name), *args, **module_args)

    def __getitem__(self, name):
        """Access items like ordinary dict."""
        return self.config[name]

    def get_logger(self, name, verbosity=2):
        msg_verbosity = "verbosity option {} is invalid. Valid options are {}.".format(
            verbosity, self.log_levels.keys()
        )
        assert verbosity in self.log_levels, msg_verbosity
        logger = logging.getLogger(name)
        logger.setLevel(self.log_levels[verbosity])
        return logger

    # setting read-only attributes
    @property
    def config(self):
        return self._config

    @property
    def save_dir(self):
        return self._save_dir

    @property
    def log_dir(self):
        return self._log_dir


# helper functions to update config dict with custom cli options
def _update_config(config, modification):
    if modification is None:
        return config

    for k, v in modification.items():
        if v is not None:
            _set_by_path(config, k, v)
    return config


def _get_opt_name(flags):
    for flg in flags:
        if flg.startswith("--"):
            return flg.replace("--", "")
    return flags[0].replace("--", "")


def _set_by_path(tree, keys, value):
    """Set a value in a nested object in tree by sequence of keys."""
    keys = keys.split(";")
    _get_by_path(tree, keys[:-1])[keys[-1]] = value


def _get_by_path(tree, keys):
    """Access a nested object in tree by sequence of keys."""
    return reduce(getitem, keys, tree)



# the GIN layer implementation adopted from Dwivedi, Vijay Prakash, et al.
# "Benchmarking graph neural networks." arXiv preprint arXiv:2003.00982 (2020).
# Github implementation, https://github.com/graphdeeplearning/benchmarking-gnns/blob/master/layers/gin_layer.py
import torch
import torch.nn as nn
import torch.nn.functional as F
from deepvelo.base import BaseModel
import dgl.function as fn

"""
    GIN: Graph Isomorphism Networks
    HOW POWERFUL ARE GRAPH NEURAL NETWORKS? (Keyulu Xu, Weihua Hu, Jure Leskovec and Stefanie Jegelka, ICLR 2019)
    https://arxiv.org/pdf/1810.00826.pdf
"""


class GINLayer(BaseModel):
    """
    [!] code adapted from dgl implementation of GINConv
    Parameters
    ----------
    apply_func : callable activation function/layer or None
        If not None, apply this function to the updated node feature,
        the :math:`f_\Theta` in the formula.
    aggr_type :
        Aggregator type to use (``sum``, ``max`` or ``mean``).
    out_dim :
        Rquired for batch norm layer; should match out_dim of apply_func if not None.
    dropout :
        Required for dropout of output features.
    batch_norm :
        boolean flag for batch_norm layer.
    residual :
        boolean flag for using residual connection.
    init_eps : optional
        Initial :math:`\epsilon` value, default: ``0``.
    learn_eps : bool, optional
        If True, :math:`\epsilon` will be a learnable parameter.

    """

    def __init__(
        self,
        apply_func,
        aggr_type,
        dropout,
        batch_norm,
        residual=False,
        init_eps=0,
        learn_eps=False,
    ):
        super().__init__()
        self.apply_func = apply_func

        if aggr_type == "sum":
            self._reducer = fn.sum
        elif aggr_type == "max":
            self._reducer = fn.max
        elif aggr_type == "mean":
            self._reducer = fn.mean
        else:
            raise KeyError("Aggregator type {} not recognized.".format(aggr_type))

        self.batch_norm = batch_norm
        self.residual = residual
        self.dropout = dropout

        in_dim = apply_func.mlp.input_dim
        out_dim = apply_func.mlp.output_dim

        if in_dim != out_dim:
            self.residual = False

        # to specify whether eps is trainable or not.
        if learn_eps:
            self.eps = torch.nn.Parameter(torch.FloatTensor([init_eps]))
        else:
            self.register_buffer("eps", torch.FloatTensor([init_eps]))

        self.bn_node_h = nn.BatchNorm1d(out_dim)

    def forward(self, g, h):
        h_in = h  # for residual connection

        g = g.local_var()
        g.ndata["h"] = h
        g.update_all(fn.copy_u("h", "m"), self._reducer("m", "neigh"))
        h = (1 + self.eps) * h + g.ndata["neigh"]
        if self.apply_func is not None:
            h = self.apply_func(h)

        if self.batch_norm:
            h = self.bn_node_h(h)  # batch normalization

        h = F.relu(h)  # non-linear activation

        if self.residual:
            h = h_in + h  # residual connection

        h = F.dropout(h, self.dropout, training=self.training)

        return h


class ApplyNodeFunc(BaseModel):
    """
    This class is used in class GINNet
    Update the node feature hv with MLP
    """

    def __init__(self, mlp):
        super().__init__()
        self.mlp = mlp

    def forward(self, h):
        h = self.mlp(h)
        return h


class MLP(BaseModel):
    """MLP with linear output"""

    def __init__(self, num_layers, input_dim, hidden_dim, output_dim):

        super().__init__()
        self.linear_or_not = True  # default is linear model
        self.num_layers = num_layers
        self.output_dim = output_dim
        self.input_dim = input_dim

        if num_layers < 1:
            raise ValueError("number of layers should be positive!")
        elif num_layers == 1:
            # Linear model
            self.linear = nn.Linear(input_dim, output_dim)
        else:
            # Multi-layer model
            self.linear_or_not = False
            self.linears = torch.nn.ModuleList()
            self.batch_norms = torch.nn.ModuleList()

            self.linears.append(nn.Linear(input_dim, hidden_dim))
            for layer in range(num_layers - 2):
                self.linears.append(nn.Linear(hidden_dim, hidden_dim))
            self.linears.append(nn.Linear(hidden_dim, output_dim))

            for layer in range(num_layers - 1):
                self.batch_norms.append(nn.BatchNorm1d((hidden_dim)))

    def forward(self, x):
        if self.linear_or_not:
            # If linear model
            return self.linear(x)
        else:
            # If MLP
            h = x
            for i in range(self.num_layers - 1):
                h = F.relu(self.batch_norms[i](self.linears[i](h)))
            return self.linears[-1](h)



from distutils.command.config import config
import numpy as np
from numpy import inf
import torch
import dgl

# from torchvision.utils import make_grid
from deepvelo.base import BaseTrainer
from deepvelo.utils import inf_loop, MetricTracker
from deepvelo.logger import TensorboardWriter


class Trainer(BaseTrainer):
    """
    Trainer class
    """

    def __init__(
        self,
        model,
        criterion,
        metric_ftns,
        optimizer,
        config,
        data_loader,
        valid_data_loader=None,
        lr_scheduler=None,
        len_epoch=None,
    ):
        super().__init__(model, criterion, metric_ftns, optimizer, config)
        self.data_loader = data_loader
        if len_epoch is None:
            # epoch-based training
            self.len_epoch = len(self.data_loader)
        else:
            # iteration-based training
            self.data_loader = inf_loop(data_loader)
            self.len_epoch = len_epoch
        self.valid_data_loader = valid_data_loader
        self.do_validation = self.valid_data_loader is not None
        self.lr_scheduler = lr_scheduler
        self.log_step = int(np.sqrt(data_loader.batch_size))

        self.train_metrics = MetricTracker(
            "loss", *[m.__name__ for m in self.metric_ftns], writer=self.writer
        )
        self.valid_metrics = MetricTracker(
            "loss", *[m.__name__ for m in self.metric_ftns], writer=self.writer
        )

    def _compute_core(self, batch_data):
        data_dict = batch_data
        x_u, x_s, target = data_dict["Ux_sz"], data_dict["Sx_sz"], data_dict["velo"]
        x_u, x_s, target = (
            x_u.to(self.device),
            x_s.to(self.device),
            target.to(self.device),
        )

        if self.config["arch"]["args"]["pred_unspliced"]:
            target_u = data_dict["velo_u"]
            target_u = target_u.to(self.device)
            # concate target to (batch, 2*genes), be careful of the order
            target = torch.cat([target, target_u], dim=1)

        output = self.model(x_u, x_s)
        return output, target

    def _smooth_constraint_step(self):
        topG = self.config["data_loader"]["args"]["topG"]
        neighbor_batch_ind = self.data_loader.dataset.gen_neighbor_batch(
            size=int(self.config["data_loader"]["args"]["batch_size"] / topG)
        )
        x_u = self.data_loader.dataset.Ux_sz[neighbor_batch_ind].to(
            self.device
        )  # (batch, genes)
        x_s = self.data_loader.dataset.Sx_sz[neighbor_batch_ind].to(self.device)
        output_neighbor = self.model(x_u, x_s)  # (batch, genes)
        self.optimizer.zero_grad()
        output_neighbor = output_neighbor.t().reshape([-1, topG])
        label_neighors = output_neighbor[:, 1:].detach()
        pivot = output_neighbor[:, 0:1]
        loss_c = torch.mean((label_neighors - pivot) ** 2)
        loss_c.backward()
        self.optimizer.step()

    def _train_epoch(self, epoch):
        """
        Training logic for an epoch

        :param epoch: Integer, current training epoch.
        :return: A log that contains average loss and metric in this epoch.
        """
        self.model.train()
        self.train_metrics.reset()
        if (not self.data_loader.shuffle) and self.data_loader.is_large_batch:
            loader = self.data_loader.dataset.large_batch(self.device)
        else:
            loader = self.data_loader
        for batch_idx, batch_data in enumerate(loader):
            output, target = self._compute_core(batch_data)
            if self.config["mask_zeros"]:
                data_dict = batch_data
                mask = data_dict["mask"].to(self.device)  # (batch, n_gene)
                output = output * mask
                target = target * mask
            self.optimizer.zero_grad()
            loss = (
                self.criterion(output, target)
                if "mle" not in self.config["loss"]["type"]
                else self.criterion(
                    output,
                    current_state=(
                        torch.cat(
                            [
                                batch_data["Sx_sz"],
                                batch_data["Ux_sz"],
                            ],
                            dim=1,
                        ).to(self.device)
                        if self.config["arch"]["args"]["pred_unspliced"]
                        else batch_data["Sx_sz"].to(self.device)
                    ),
                    idx=batch_data["t+1 neighbor idx"].to(self.device),
                    candidate_states=self.candidate_states,
                    spliced_counts=batch_data["Sx_sz"].to(self.device),
                    unspliced_counts=batch_data["Ux_sz"].to(self.device),
                    **self.config["loss"]["args"],
                )
            )
            loss.backward()
            self.optimizer.step()

            if self.config["constraint_loss"]:
                self._smooth_constraint_step()

            self.writer.set_step((epoch - 1) * self.len_epoch + batch_idx)
            self.train_metrics.update("loss", loss.item())
            for met in self.metric_ftns:
                self.train_metrics.update(met.__name__, met(output, target))

            if batch_idx % self.log_step == 0:
                self.logger.debug(
                    "Train Epoch: {} {} Loss: {:.6f}".format(
                        epoch, self._progress(batch_idx), loss.item()
                    )
                )
                # self.writer.add_image('input', make_grid(data.cpu(), nrow=8, normalize=True))

            if batch_idx == self.len_epoch:
                break
        log = self.train_metrics.result()

        if self.do_validation:
            val_log = self._valid_epoch(epoch)
            log.update(**{"val_" + k: v for k, v in val_log.items()})

        if self.lr_scheduler is not None:
            self.lr_scheduler.step()
        return log

    def _valid_epoch(self, epoch):
        """
        Validate after training an epoch

        :param epoch: Integer, current training epoch.
        :return: A log that contains information about validation
        """
        self.model.eval()
        self.valid_metrics.reset()
        with torch.no_grad():
            for batch_idx, batch_data in enumerate(self.valid_data_loader):
                output, target = self._compute_core(batch_data)
                loss = self.criterion(output, target)

                self.writer.set_step(
                    (epoch - 1) * len(self.valid_data_loader) + batch_idx, "valid"
                )
                self.valid_metrics.update("loss", loss.item())
                for met in self.metric_ftns:
                    self.valid_metrics.update(met.__name__, met(output, target))
                # self.writer.add_image('input', make_grid(data.cpu(), nrow=8, normalize=True))

        # add histogram of model parameters to the tensorboard
        for name, p in self.model.named_parameters():
            self.writer.add_histogram(name, p, bins="auto")
        return self.valid_metrics.result()

    def _progress(self, batch_idx):
        base = "[{}/{} ({:.0f}%)]"
        if hasattr(self.data_loader, "n_samples"):
            current = batch_idx * self.data_loader.batch_size
            total = self.data_loader.n_samples
        else:
            current = batch_idx
            total = self.len_epoch
        return base.format(current, total, 100.0 * current / total)

    def eval(self, eval_loader, return_kinetic_rates=False):
        """
        Evaluate the model on a given dataset, provided by eval_loader

        :param model: model to evaluate
        :param eval_loader: dataset loader containing dataset to evaluate on
        """
        self.model.eval()
        n_genes = self.config["arch"]["args"]["n_genes"]
        velo_mat = []
        velo_mat_u = []
        kinetic_rates = {}
        if (not eval_loader.shuffle) and eval_loader.is_large_batch:
            loader = eval_loader.dataset.large_batch(self.device)
        else:
            loader = eval_loader
        with torch.no_grad():
            for batch_idx, batch_data in enumerate(loader):
                output, _ = self._compute_core(batch_data)
                pred_s = output[:, 0:n_genes]
                velo_mat.append(pred_s.cpu().data)
                if self.config["arch"]["args"]["pred_unspliced"]:
                    pred_u = output[:, n_genes:]
                    velo_mat_u.append(pred_u.cpu().data)

                # record kinetic rates
                if return_kinetic_rates:
                    cur_kinetic_rates = self.model.get_current_batch_kinetic_rates()
                    for k, v in cur_kinetic_rates.items():
                        if k not in kinetic_rates:
                            kinetic_rates[k] = []
                        kinetic_rates[k].append(v.cpu().data)

            velo_mat = np.concatenate(velo_mat, axis=0)
            if return_kinetic_rates:
                for k, v in kinetic_rates.items():
                    kinetic_rates[k] = np.concatenate(v, axis=0)
            if self.config["arch"]["args"]["pred_unspliced"]:
                velo_mat_u = np.concatenate(velo_mat_u, axis=0)
        return velo_mat, velo_mat_u, kinetic_rates



from typing import Mapping, Optional, Tuple

import torch
import numpy as np
import torch.nn.functional as F
from sklearn import neighbors


def nll_loss(output, target):
    return F.nll_loss(output, target)


def mse_loss(output, target):
    return F.mse_loss(output, target)


def min_loss(output, target, alpha=0, mode="per gene"):
    """
    selects one closest cell and computes the loss

    the target is the set of velocity target candidates,
    find the closest in them.

    output: torch.tensor e.g. (128, 2000)
    target: torch.tensor e.g. (128, 30, 2000)
    """
    distance = torch.pow(
        target - torch.unsqueeze(output, 1), exponent=2
    )  # (128, 30, 2000)
    if mode == "per gene":
        distance = torch.min(distance, dim=1)[
            0
        ]  # find the closest target for each gene (128, 2000)
        min_distance = torch.sum(distance, dim=1)  # (128,)
    elif mode == "per cell":
        distance = torch.sum(distance, dim=2)  # (128, 30)
        min_distance = torch.min(distance, dim=1)[0]  # (128,)
    else:
        raise NotImplementedError

    # loss = torch.mean(torch.max(torch.tensor(alpha).float(), min_distance))
    loss = torch.mean(min_distance)

    return loss


def _find_candidates(
    output: torch.Tensor,
    current_state: torch.Tensor,
    idx: torch.LongTensor,
    candidate_states: torch.Tensor,
    n_genes: int,
    inner_batch_size: int,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Find the candidate states for each cell in the batch.

    Args:
    output (torch.Tensor):
        The predicted velocity vectors, (batch_size, genes).
    current_state (torch.Tensor):
        The current cell state, (batch_size, genes).
    idx (torch.LongTensor):
        The indices future nearest neighbors, (batch_size, num_neighbors).
    candidate_states (torch.Tensor):
        The states of potential future nearest neighbors, (all_cells, genes).
    n_genes (int):
        The number of spliced genes.

    Returns:
    delta_tp1 (torch.Tensor):
        The predicted average difference to t+1 state, (batch_size, genes).
    delta_tm1 (torch.Tensor):
        The predicted average difference to t-1 state, (batch_size, genes).
    """
    inner_batch_size = min(inner_batch_size, output.shape[0])
    num_neighbors = idx.shape[1]
    delta_tp1 = []
    delta_tm1 = []
    for i in range(0, output.shape[0], inner_batch_size):
        batch_output = output[i : i + inner_batch_size]  # (inner_batch, genes)
        batch_current_state = current_state[
            i : i + inner_batch_size
        ]  # (inner_batch, genes)
        batch_idx = idx[i : i + inner_batch_size]  # (inner_batch, num_neighbors)
        flatten_batch_idx = batch_idx.view(-1)  # (inner_batch * num_neighbors)
        flatten_candidate_state = candidate_states[
            flatten_batch_idx, :
        ]  # (inner_batch * num_neighbors, genes)
        candidate_state = flatten_candidate_state.view(
            batch_idx.shape[0], num_neighbors, -1
        )
        delta = candidate_state - batch_current_state.unsqueeze(
            1
        )  # (inner_batch, num_neighbors, genes)
        cos_sim = F.cosine_similarity(
            delta[:, :, :n_genes],
            batch_output[:, :n_genes].unsqueeze(1),
            dim=2,
        )  # (inner_batch, num_neighbors)

        candidates_p1 = cos_sim > 0  # (inner_batch, num_neighbors)
        num_p1 = torch.sum(candidates_p1, dim=1)  # (inner_batch)
        batch_delta_p1 = torch.sum(delta * candidates_p1.unsqueeze(2).float(), dim=1)
        batch_delta_p1 = batch_delta_p1 / (
            num_p1.unsqueeze(1).float() + 1e-9
        )  # (inner_batch, genes)
        delta_tp1.append(batch_delta_p1)

        candidates_m1 = cos_sim < 0  # (inner_batch, num_neighbors)
        num_m1 = torch.sum(candidates_m1, dim=1)  # (inner_batch)
        batch_delta_m1 = torch.sum(delta * candidates_m1.unsqueeze(2).float(), dim=1)
        batch_delta_m1 = batch_delta_m1 / (
            num_m1.unsqueeze(1).float() + 1e-9
        )  # (inner_batch, genes)
        delta_tm1.append(batch_delta_m1)

    delta_tp1 = torch.cat(delta_tp1, dim=0)  # (batch_size, genes)
    delta_tm1 = torch.cat(delta_tm1, dim=0)  # (batch_size, genes)
    return delta_tp1, delta_tm1


def integrate_mle(
    output: torch.Tensor,
    current_state: torch.Tensor,
    idx: torch.LongTensor,
    candidate_states: torch.Tensor,
    n_spliced: int = None,
    *args,
    **kwargs,
) -> torch.Tensor:
    """
    The specialized maximum likelihood estimation loss for predicting possible
    future cell states.

    Args:
    output (torch.Tensor):
        The predicted velocity vectors, (batch_size, genes).
    current_state (torch.Tensor):
        The current cell state, (batch_size, genes).
    idx (torch.LongTensor):
        The indices future nearest neighbors, (batch_size, num_neighbors).
    candidate_states (torch.Tensor):
        The states of potential future nearest neighbors, (all_cells, genes).

    Returns: (torch.Tensor)
    """
    batch_size, genes = current_state.shape
    if n_spliced is not None:
        genes = n_spliced
    with torch.no_grad():
        delta_tp1, delta_tm1 = _find_candidates(
            output,
            current_state,
            idx,
            candidate_states,
            n_genes=genes,
            inner_batch_size=int(5e8 / idx.shape[1] / current_state.shape[1]),
        )  # (batch_size, genes)
    loss_tp1 = torch.mean(torch.pow(output - delta_tp1, 2))
    loss_tm1 = torch.mean(torch.pow(output + delta_tm1, 2))
    loss = (loss_tp1 + loss_tm1) / 2 * np.sqrt(genes)
    return loss


def mle(
    output: torch.Tensor,
    current_state: torch.Tensor,
    idx: torch.LongTensor,
    candidate_states: torch.Tensor,
    n_spliced: int = None,
    *args,
    **kwargs,
) -> torch.Tensor:
    """
    The specialized maximum likelihood estimation loss for predicting possible
    future cell states. The implementation loops over the neighbor dimension,
    which is different from the equation in paper but works the same way. This
    has higher computational efficiency.

    Args:
    output (torch.Tensor):
        The predicted velocity vectors, (batch_size, genes).
    current_state (torch.Tensor):
        The current cell state, (batch_size, genes).
    idx (torch.LongTensor):
        The indices future nearest neighbors, (batch_size, num_neighbors).
    candidate_states (torch.Tensor):
        The states of potential future nearest neighbors, (all_cells, genes).

    Returns: (torch.Tensor)
    """
    batch_size, genes = current_state.shape
    if n_spliced is not None:
        genes = n_spliced
    num_neighbors = idx.shape[1]
    loss = []
    for i in range(num_neighbors):
        ith_neighbors = idx[:, i]
        candidate_state = candidate_states[ith_neighbors, :]  # (batch_size, genes)
        delta = (candidate_state - current_state).detach()  # (batch_size, genes)
        cos_sim = F.cosine_similarity(
            delta[:, :genes], output[:, :genes], dim=1
        )  # (batch_size,)

        # t+1 direction
        candidates = cos_sim.detach() > 0
        squared_difference = torch.mean(torch.pow(output - delta, 2), dim=1)
        squared_difference = squared_difference[candidates]
        squared_difference = (
            squared_difference / torch.mean(torch.pow(delta, 2)) * np.sqrt(genes)
        )  # scaling
        loss.append(torch.sum(squared_difference) / len(candidates))

        # t-1 direction, (-output) - delta
        candidates = cos_sim.detach() < 0
        squared_difference = torch.mean(torch.pow(output + delta, 2), dim=1)
        squared_difference = squared_difference[candidates]
        squared_difference = (
            squared_difference / torch.mean(torch.pow(delta, 2)) * np.sqrt(genes)
        )  # scaling
        loss.append(torch.sum(squared_difference) / len(candidates))
        # TODO: check why the memory usage is high for this one
    loss = torch.stack(loss).mean()
    return loss


def pearson(
    x: torch.Tensor,
    y: torch.Tensor,
    mask: Optional[torch.BoolTensor] = None,
) -> torch.Tensor:
    """
    The pearson correlation coefficient between two vectors. by default, the
    first dimension is the sample dimension.

    Args:
    x (torch.Tensor):
        The first vector, (N, D).
    y (torch.Tensor):
        The second vector, (N, D).
    mask (torch.BoolTensor):
        The mask showing valid data positions, (N, D).

    Returns: (torch.Tensor), (D,)
    """
    if mask is None:
        x = x - torch.mean(x, dim=0)
        y = y - torch.mean(y, dim=0)
        x = x / (torch.std(x, dim=0) + 1e-9)
        y = y / (torch.std(y, dim=0) + 1e-9)
        return torch.mean(x * y, dim=0)  # (D,)
    else:
        assert mask.dtype == torch.bool
        mask = mask.detach().float()
        num_valid_data = torch.sum(mask, dim=0)  # (D,)

        y = y * mask
        x = x * mask
        x = x - torch.sum(x, dim=0) / (num_valid_data + 1e-9)
        y = y - torch.sum(y, dim=0) / (num_valid_data + 1e-9)
        y = y * mask  # make the invalid data to zero again to ignore
        x = x * mask
        x = x / (torch.sqrt(torch.sum(torch.pow(x, 2), dim=0)) + 1e-9)
        y = y / (torch.sqrt(torch.sum(torch.pow(y, 2), dim=0)) + 1e-9)
        return torch.sum(x * y, dim=0)  # (D,)


def direction_loss(
    velocity: torch.Tensor,
    spliced_counts: torch.Tensor,
    unspliced_counts: torch.Tensor,
    velocity_u: Optional[torch.Tensor] = None,
    coeff_u: float = 1.0,
    coeff_s: float = 1.0,
) -> torch.Tensor:
    """
    The constraint for the direction of the velocity vectors. Large ratio of u/s
    should have positive direction, for each gene.

    Args:
    velocity (torch.Tensor):
        The predicted velocity vectors, (batch_size, genes).
    spliced_counts (torch.Tensor):
        The number of spliced reads, (batch_size, genes).
    unspliced_counts (torch.Tensor):
        The number of unspliced reads, (batch_size, genes).
    velocity_u (torch.Tensor):
        The predicted velocity vectors for unspliced reads, (batch_size, genes).

    Returns: (torch.Tensor), (1,)
    """

    # 3. Intereting for the gliogenic cells of the GABAInterneuraons ====
    mask1 = unspliced_counts > 0
    mask2 = spliced_counts > 0
    corr = coeff_u * pearson(velocity, unspliced_counts, mask1) - coeff_s * pearson(
        velocity, spliced_counts, mask2
    )

    # 4.===============================================================
    # mask = (spliced_counts > 0) & (unspliced_counts > 0)
    # corr = coeff_u * pearson(velocity, unspliced_counts, mask) - coeff_s * pearson(
    #     velocity, spliced_counts, mask
    # )
    loss = coeff_u + coeff_s - torch.mean(corr)  # to maximize the correlation
    loss = loss / (coeff_u + coeff_s)
    # should not use correlation to u, since the alpha coefficient is unknown and
    # it definitely varies along the phase portrait.
    # if velocity_u is not None:
    #     corr_unpliced = pearson(velocity_u, unspliced_counts, mask)
    #     loss_u = 1 + torch.mean(corr_unpliced)  # mininize the corr_unpliced
    #     loss = 0.5 * (loss + loss_u)
    return loss  # [0, 2.0]


def mle_plus_direction(
    output: torch.Tensor,
    current_state: torch.Tensor,
    idx: torch.LongTensor,
    candidate_states: torch.Tensor,
    spliced_counts: torch.Tensor,
    unspliced_counts: torch.Tensor,
    pearson_scale: float = 10.0,
    coeff_u: float = 1.0,
    coeff_s: float = 1.0,
) -> torch.Tensor:
    """
    The combination of maximum likelihood estimation loss and direction loss.
    """
    batch_size, genes = current_state.shape
    if output.shape[1] == spliced_counts.shape[1]:
        velocity = output
    elif output.shape[1] == 2 * spliced_counts.shape[1]:
        # when predicting unspliced velocity, the output demension equals 2*genes
        velocity = output[:, : spliced_counts.shape[1]]
    else:
        raise ValueError(
            "output dimension is not correct, "
            "it should either be num of genes when `pred_unspliced` is False, "
            "or 2 * num of genes when `pred_unspliced` is True."
        )

    loss_mle = integrate_mle(
        output,
        current_state,
        idx,
        candidate_states,
        n_spliced=velocity.shape[1],
    )
    loss_pearson = direction_loss(
        velocity,
        spliced_counts,
        unspliced_counts,
        coeff_u=coeff_u,
        coeff_s=coeff_s,
    )

    # scale the pearson to make it comparable with mle
    # TODO: should scale by * np.sqrt(genes)? and negative scale to batch size?
    _lambda = (pearson_scale * np.sqrt(genes)) / min(
        0.5, max(0.01, 1 - loss_pearson.item())
    )
    # print(f"mle: {loss_mle.item()}, direction_loss: {loss_pearson.item()}")
    return loss_mle + _lambda * loss_pearson


def sim_loss():
    """the loss measuring similarities among input vectors"""
    return



from typing import Optional, Tuple

import numpy as np
import scvelo as scv
from scvelo import logging as logg
from scvelo.core import sum as sum_
from anndata import AnnData

from deepvelo.utils.plot import dist_plot


def clip_and_norm_Ms_Mu(
    adata,
    do_clip: bool = True,
    do_norm: bool = True,
    target_mean: float = 0.4,
    replace: bool = False,
    save_fig: Optional[str] = None,
    plot: bool = True,
    print_summary: bool = True,
) -> Tuple[float, float]:
    """
    Normalize using the mean and standard deviation of the gene expression matrix.

    Args:
        adata (Anndata): Anndata object.
        target_mean (float): target mean.
        replace (bool): replace the original data.
        save_fig (str): directory to save figures.
        plot (bool): plot the distribution of the normalized data.
        print_summary (bool): print the summary of the normalized data.

    Returns:
        Tupel[float, float]: scale factor for Ms and Mu.
    """
    non_zero_Ms = adata.layers["Ms"][adata.layers["Ms"] > 0]
    non_zero_Mu = adata.layers["Mu"][adata.layers["Mu"] > 0]
    if print_summary:
        print(
            f"Raw Ms: mean {adata.layers['Ms'].mean():.2f},"
            f" max {adata.layers['Ms'].max():.2f},"
            f" std {adata.layers['Ms'].std():.2f},"
            f" 99.5% quantile {np.percentile(adata.layers['Ms'], 99.5):.2f}"
            f" 99.5% of non-zero: {np.percentile(non_zero_Ms, 99.5):.2f}"
        )
        print(
            f"Raw Mu: mean {adata.layers['Mu'].mean():.2f},"
            f" max {adata.layers['Mu'].max():.2f},"
            f" std {adata.layers['Mu'].std():.2f},"
            f" 99.5% quantile {np.percentile(adata.layers['Mu'], 99.5):.2f}"
            f" 99.5% of non-zero: {np.percentile(non_zero_Mu, 99.5):.2f}"
        )

    if do_clip:
        # clip the max value to 99.5% quantile
        adata.layers["NMs"] = np.clip(
            adata.layers["Ms"], None, np.percentile(non_zero_Ms, 99.5)
        )
        adata.layers["NMu"] = np.clip(
            adata.layers["Mu"], None, np.percentile(non_zero_Mu, 99.5)
        )
    else:
        adata.layers["NMs"] = adata.layers["Ms"]
        adata.layers["NMu"] = adata.layers["Mu"]
    logg.hint(f"added 'NMs' (adata.layers)")
    logg.hint(f"added 'NMu' (adata.layers)")

    if plot:
        dist_plot(
            adata.layers["NMs"].flatten(),
            adata.layers["NMu"].flatten(),
            bins=20,
            labels=["NMs", "NMu"],
            title="Distribution of Ms and Mu",
            save=f"{save_fig}/hist-Ms-Mu.png" if save_fig is not None else None,
        )

    scale_Ms, scale_Mu = 1.0, 1.0
    if do_norm:
        scale_Ms = adata.layers["NMs"].mean() / target_mean
        scale_Mu = adata.layers["NMu"].mean() / target_mean
        adata.layers["NMs"] = adata.layers["NMs"] / scale_Ms
        adata.layers["NMu"] = adata.layers["NMu"] / scale_Mu
        print(f"Normalized Ms and Mu to mean of {target_mean}")
        if plot:
            ax = scv.pl.hist(
                [adata.layers["NMs"].flatten(), adata.layers["NMu"].flatten()],
                labels=["NMs", "NMu"],
                kde=False,
                normed=False,
                bins=20,
                # xlim=[0, 1],
                fontsize=18,
                legend_fontsize=16,
                show=False,
            )
            if save_fig is not None:
                ax.get_figure().savefig(f"{save_fig}/hist-normed-Ms-Mu.png")

    if print_summary:
        print(
            f"New Ms: mean {adata.layers['NMs'].mean():.2f},"
            f" max {adata.layers['NMs'].max():.2f},"
            f" std {adata.layers['NMs'].std():.2f},"
            f" 99.5% quantile {np.percentile(adata.layers['NMs'], 99.5):.2f}"
        )
        print(
            f"New Mu: mean {adata.layers['NMu'].mean():.2f},"
            f" max {adata.layers['NMu'].max():.2f},"
            f" std {adata.layers['NMu'].std():.2f},"
            f" 99.5% quantile {np.percentile(adata.layers['NMu'], 99.5):.2f}"
        )

    if replace:
        adata.layers["Ms"] = adata.layers["NMs"]
        adata.layers["Mu"] = adata.layers["NMu"]
        logg.hint(f"replaced 'Ms' (adata.layers) with 'NMs'")
        logg.hint(f"replaced 'Mu' (adata.layers) with 'NMu'")

    return scale_Ms, scale_Mu


def autoset_coeff_s(adata: AnnData, use_raw: bool = True) -> float:
    """
    Automatically set the weighting for objective term of the spliced
    read correlation. Modified from the scv.pl.proportions function.

    Args:
        adata (Anndata): Anndata object.
        use_raw (bool): use raw data or processed data.

    Returns:
        float: weighting coefficient for objective term of the unpliced read
    """
    layers = ["spliced", "unspliced", "ambigious"]
    layers_keys = [key for key in layers if key in adata.layers.keys()]
    counts_layers = [sum_(adata.layers[key], axis=1) for key in layers_keys]

    if use_raw:
        ikey, obs = "initial_size_", adata.obs
        counts_layers = [
            obs[ikey + layer_key] if ikey + layer_key in obs.keys() else c
            for layer_key, c in zip(layers_keys, counts_layers)
        ]
    counts_total = np.sum(counts_layers, 0)
    counts_total += counts_total == 0
    counts_layers = np.array([counts / counts_total for counts in counts_layers])
    counts_layers = np.mean(counts_layers, axis=1)

    spliced_counts = counts_layers[layers_keys.index("spliced")]
    ratio = spliced_counts / counts_layers.sum()

    if ratio < 0.7:
        coeff_s = 0.5
        print(
            f"The ratio of spliced reads is {ratio*100:.1f}% (less than 70%). "
            f"Suggest using coeff_s {coeff_s}."
        )
    elif ratio < 0.85:
        coeff_s = 0.75
        print(
            f"The ratio of spliced reads is {ratio*100:.1f}% (between 70% and 85%). "
            f"Suggest using coeff_s {coeff_s}."
        )
    else:
        coeff_s = 1.0
        print(
            f"The ratio of spliced reads is {ratio*100:.1f}% (more than 85%). "
            f"Suggest using coeff_s {coeff_s}."
        )

    return coeff_s



from copy import deepcopy
from typing import Callable, Mapping
import numpy as np

import torch
from deepvelo.trainer import Trainer
from scvelo import logging as logg

import deepvelo.data_loader.data_loaders as module_data
import deepvelo.model.loss as module_loss
import deepvelo.model.metric as module_metric
import deepvelo.model.model as module_arch
from deepvelo.parse_config import ConfigParser

# a hack to make constants, see https://stackoverflow.com/questions/3203286
class MetaConstants(type):
    @property
    def default_configs(cls):
        return deepcopy(cls._default_configs)


class Constants(object, metaclass=MetaConstants):
    _default_configs = {
        "name": "DeepVelo_Base",
        "n_gpu": 1,  # whether to use GPU
        "arch": {
            "type": "VeloGCN",
            "args": {
                "layers": [64, 64],
                "dropout": 0.2,
                "fc_layer": False,
                "pred_unspliced": False,
            },
        },
        "data_loader": {
            "type": "VeloDataLoader",
            "args": {
                "shuffle": False,
                "validation_split": 0.0,
                "num_workers": 2,
                "type": "pca, t",
                "topC": 30,
                "topG": 20,
            },
        },
        "optimizer": {
            "type": "Adam",
            "args": {"lr": 0.001, "weight_decay": 0, "amsgrad": True},
        },
        "loss": {
            "type": "mle_plus_direction",
            "args": {
                "pearson_scale": 18.0,
                "coeff_u": 1.0,
                "coeff_s": 1.0,
            },
        },
        "constraint_loss": False,
        "mask_zeros": False,
        "metrics": ["mse"],
        "lr_scheduler": {"type": "StepLR", "args": {"step_size": 1, "gamma": 0.97}},
        "trainer": {
            "epochs": 100,
            "save_dir": "saved/",
            "save_period": 1000,
            "verbosity": 1,
            "monitor": "min mse",
            "early_stop": 1000,
            "tensorboard": True,
        },
    }


def train(
    adata,
    configs: Mapping,
    verbose: bool = False,
    return_kinetic_rates: bool = True,
    callback: Callable = None,
    **kwargs,
):
    batch_size, n_genes = adata.layers["Ms"].shape
    configs["arch"]["args"]["n_genes"] = n_genes
    configs["data_loader"]["args"]["batch_size"] = batch_size
    config = ConfigParser(configs)
    logger = config.get_logger("train")

    # setup data_loader instances, use adata as the data_source to load inmemory data
    data_loader = config.init_obj("data_loader", module_data, data_source=adata)
    valid_data_loader = data_loader.split_validation()

    # build model architecture, then print to console
    if config["arch"]["type"] in ["VeloGCN", "VeloGIN"]:
        model = config.init_obj("arch", module_arch, g=data_loader.dataset.g)
    else:
        model = config.init_obj("arch", module_arch)
    logger.info(f"Beginning training of {configs['name']} ...")
    if verbose:
        logger.info(configs)
        logger.info(model)

    # get function handles of loss and metrics
    criterion = getattr(module_loss, configs["loss"]["type"])
    metrics = [getattr(module_metric, met) for met in configs["metrics"]]

    # build optimizer, learning rate scheduler. delete every lines containing lr_scheduler for disabling scheduler
    trainable_params = filter(lambda p: p.requires_grad, model.parameters())
    optimizer = config.init_obj("optimizer", torch.optim, trainable_params)
    lr_scheduler = config.init_obj("lr_scheduler", torch.optim.lr_scheduler, optimizer)

    trainer = Trainer(
        model,
        criterion,
        metrics,
        optimizer,
        config=config,
        data_loader=data_loader,
        valid_data_loader=valid_data_loader,
        lr_scheduler=lr_scheduler,
    )

    def callback_wrapper(epoch):
        # evaluate all and return the velocity matrix (cells, features)
        config_copy = configs["data_loader"]["args"].copy()
        config_copy.update(shuffle=False, training=False, data_source=adata)
        eval_loader = getattr(module_data, configs["data_loader"]["type"])(
            **config_copy
        )
        velo_mat, velo_mat_u, kinetic_rates = trainer.eval(
            eval_loader, return_kinetic_rates=return_kinetic_rates
        )

        if callback is not None:
            callback(adata, velo_mat, velo_mat_u, kinetic_rates, epoch)
        else:
            logg.warn(
                "Set verbose to True but no callback function provided. A possible "
                "callback function accepts at least two arguments: adata, velo_mat "
            )

    if verbose:
        trainer.train_with_epoch_callback(
            callback=callback_wrapper,
            freq=kwargs.get("freq", 30),
        )
    else:
        trainer.train()

    if configs["data_loader"]["args"]["shuffle"] == False:
        eval_loader = data_loader
    else:
        config_copy = configs["data_loader"]["args"].copy()
        config_copy.update(shuffle=False, training=False, data_source=adata)
        eval_loader = getattr(module_data, configs["data_loader"]["type"])(
            **config_copy
        )
    velo_mat, velo_mat_u, kinetic_rates = trainer.eval(
        eval_loader, return_kinetic_rates=return_kinetic_rates
    )

    print("velo_mat shape:", velo_mat.shape)
    # add velocity
    assert adata.layers["Ms"].shape == velo_mat.shape
    adata.layers["velocity"] = velo_mat  # (cells, genes)
    if len(velo_mat_u) > 0:
        adata.layers["velocity_unspliced"] = velo_mat_u

    logg.hint(f"added 'velocity' (adata.layers)")
    logg.hint(f"added 'velocity_unspliced' (adata.layers)")

    if return_kinetic_rates:
        for k, v in kinetic_rates.items():
            if v is not None:
                adata.layers["cell_specific_" + k] = v
                logg.hint(f"added 'cell_specific_{k}' (adata.layers)")
    return trainer



import pickle
from pathlib import Path

# from torchvision import datasets, transforms
import numpy as np
import torch
import dgl
import hnswlib
from dgl.contrib.sampling import NeighborSampler
from torch.utils.data import Dataset
from sklearn.metrics import pairwise_distances
from sklearn.decomposition import PCA
from anndata import AnnData

from deepvelo.base import BaseDataLoader


class VeloDataset(Dataset):
    def __init__(
        self,
        data_source,
        train=True,
        type="average",
        topC=30,
        topG=20,
        nn_smooth=False,
    ):
        # check if data_source is a file path or inmemory data
        if isinstance(data_source, str):
            data_source = Path(data_source)
            with open(data_source, "rb") as f:
                adata = pickle.load(f)
        elif isinstance(data_source, AnnData):
            adata = data_source
        else:
            raise ValueError("data_source must be a file path or anndata object")
        self.Ux_sz = adata.layers["Mu"]
        self.Sx_sz = adata.layers["Ms"]
        self.connectivities = adata.obsp["connectivities"]  # shape (cells, features)

        self.topG = topG
        N_cell, N_gene = self.Sx_sz.shape

        # build the knn graph in the original space
        # TODO: try using the original connectivities to build the graph
        if "pca" in type:
            n_pcas = 30
            pca_ = PCA(
                n_components=n_pcas,
                svd_solver="randomized",
            )
            Sx_sz_pca = pca_.fit_transform(self.Sx_sz)
            if N_cell < 3000:
                ori_dist = pairwise_distances(Sx_sz_pca, Sx_sz_pca)
                self.ori_idx = np.argsort(ori_dist, axis=1)[:, :topG]  # (1720, 20)
                self.nn_t_idx = np.argsort(ori_dist, axis=1)[:, 1:topC]
            else:
                p = hnswlib.Index(space="l2", dim=n_pcas)
                p.init_index(max_elements=N_cell, ef_construction=200, M=30)
                p.add_items(Sx_sz_pca)
                p.set_ef(max(topC, topG) + 10)
                self.ori_idx = p.knn_query(Sx_sz_pca, k=topG)[0].astype(int)
                self.nn_t_idx = p.knn_query(Sx_sz_pca, k=topC)[0][:, 1:].astype(int)
        else:
            raise NotImplementedError(
                "the argument type of VeloDataset has to include original "
                "distance method 'pca' or 'raw'"
            )

        self.g = self.build_graph(self.ori_idx)
        self.nn_idx = self.nn_t_idx
        self.neighbor_time = 0

        # update the velocity target vectors for spliced and unspliced counts
        self.velo = np.zeros(self.Sx_sz.shape, dtype=np.float32)
        self.velo_u = np.zeros(self.Ux_sz.shape, dtype=np.float32)
        for i in range(N_cell):
            self.velo[i] = np.mean(self.Sx_sz[self.nn_idx[i]], axis=0) - self.Sx_sz[i]
            self.velo_u[i] = np.mean(self.Ux_sz[self.nn_idx[i]], axis=0) - self.Ux_sz[i]

        # build masks
        mask = np.ones([N_cell, N_gene])
        mask[self.Ux_sz == 0] = 0
        mask[self.Sx_sz == 0] = 0

        self.Ux_sz = torch.tensor(self.Ux_sz, dtype=torch.float32)
        self.Sx_sz = torch.tensor(self.Sx_sz, dtype=torch.float32)
        self.velo = torch.tensor(self.velo, dtype=torch.float32)
        self.velo_u = torch.tensor(self.velo_u, dtype=torch.float32)
        self.mask = torch.tensor(mask, dtype=torch.float32)
        print("velo data shape:", self.velo.shape)

    def large_batch(self, device):
        """
        build the large batch for training
        """
        # check if self._large_batch is already built
        if hasattr(self, "_large_batch"):
            return self._large_batch
        self._large_batch = [
            {
                "Ux_sz": self.Ux_sz.to(device),
                "Sx_sz": self.Sx_sz.to(device),
                "velo": self.velo.to(device),
                "velo_u": self.velo_u.to(device),
                "mask": self.mask.to(device),
                "t+1 neighbor idx": torch.tensor(
                    self.nn_t_idx,
                    dtype=torch.long,
                ).to(device),
            }
        ]
        return self._large_batch

    def __len__(self):
        return len(self.Ux_sz)  # 1720

    def __getitem__(self, i):

        data_dict = {
            "Ux_sz": self.Ux_sz[i],
            "Sx_sz": self.Sx_sz[i],
            "velo": self.velo[i],
            "velo_u": self.velo_u[i],
            "mask": self.mask[i],
            "t+1 neighbor idx": self.nn_t_idx[i],
        }
        return data_dict

    def gen_neighbor_batch(self, size):
        indices = np.random.random_integers(0, high=len(self) - 1, size=size)
        # self.neighbors_per_gene is the neighbor indices for all cells, shape
        # (N_cells, topG, genes)

        # TODO(Haotian): try the per gene version
        # Here since the per gene version encounters the 0 gene count bug, we first
        # use the per cell version, which is using self.ind
        return self.ind[indices, : self.topG].flatten()

    def build_graph(self, ind):
        """ind (N,k) contains neighbor index"""
        print("building graph")
        g = dgl.DGLGraph()
        g.add_nodes(len(self.Ux_sz))
        edge_list = []
        for i in range(ind.shape[0]):
            for j in range(ind.shape[1]):
                edge_list.append((i, ind[i, j]))
        # add edges two lists of nodes: src and dst
        src, dst = tuple(zip(*edge_list))
        g.add_edges(src, dst)
        # edges are directional in DGL; make them bi-directional
        g.add_edges(dst, src)
        return g


class VeloDataLoader(BaseDataLoader):
    """
    MNIST data loading demo using BaseDataLoader
    """

    def __init__(
        self,
        data_source,
        batch_size,
        shuffle=True,
        validation_split=0.0,
        num_workers=1,
        training=True,
        type="average",
        topC=30,
        topG=16,
    ):
        self.data_source = data_source
        self.dataset = VeloDataset(
            data_source, train=training, type=type, topC=topC, topG=topG
        )
        self.shuffle = shuffle
        self.is_large_batch = batch_size == len(self.dataset)
        super().__init__(
            self.dataset, batch_size, shuffle, validation_split, num_workers
        )


class VeloNeighborSampler(NeighborSampler, BaseDataLoader):
    """
    minibatch neighbor sampler using DGL NeighborSampler
    """

    def __init__(
        self,
        data_dir,
        batch_size,
        num_neighbors,
        num_hops,
        shuffle=True,
        validation_split=0.0,
        num_workers=32,
        training=True,
    ):
        self.data_dir = data_dir
        self.dataset = VeloDataset(self.data_dir, train=training)
        # FIXME: the split_validation here is not working as in the BaseDataLoader
        # BaseDataLoader.__init__(self, self.dataset, batch_size, shuffle,
        # validation_split, num_workers)

        g = self.dataset.g
        norm = 1.0 / g.in_degrees().float().unsqueeze(1)
        g.ndata["Ux_sz"] = self.dataset.Ux_sz
        g.ndata["Sx_sz"] = self.dataset.Sx_sz
        g.ndata["velo"] = self.dataset.velo
        g.ndata["norm"] = norm
        # need to set to readonly for nodeflow
        g.readonly()

        NeighborSampler.__init__(
            self,
            g,
            batch_size,
            num_neighbors,
            neighbor_type="in",
            shuffle=shuffle,
            num_workers=num_workers,
            num_hops=num_hops,
            #  seed_nodes=train_nid
        )

    # FIXME: the split_validation here is not working as in the BaseDataLoader
    def split_validation(self):
        return None

    def __len__(self):
        return self.dataset.__len__()


if __name__ == "__main__":
    VeloDataset("./data/DG_norm_genes.npz")
    VeloNeighborSampler("./data/DG_norm_genes.npz", 32, 15, 4)



import numpy as np
from sklearn.metrics.pairwise import cosine_similarity

from scvelo import logging as logg
from scvelo.core import l2_norm, prod_sum
from scvelo.preprocessing.neighbors import get_neighs
from .util import get_indices

# Code modified from function velocity_confidence from
# https://github.com/theislab/scvelo/blob/master/scvelo/tools/velocity_confidence.py
def velocity_confidence(
    data,
    vkey="velocity",
    method="corr",
    scope_key=None,
    copy=False,
    only_velocity_genes=False,
    only_high_spearman=False,
):
    """Computes confidences of velocities.

    .. code:: python

        scv.tl.velocity_confidence(adata)
        scv.pl.scatter(adata, color='velocity_confidence', perc=[2,98])

    .. image:: https://user-images.githubusercontent.com/31883718/69626334-b6df5200-1048-11ea-9171-495845c5bc7a.png
       :width: 600px


    Arguments
    ---------
    data: :class:`~anndata.AnnData`
        Annotated data matrix.
    vkey: `str` (default: `'velocity'`)
        Name of velocity estimates to be used.
    method: `str` (default: `'corr'`), choice of `'corr'` or `'cosine'`
        Method to use for computing confidence, whether to use correlation or
        cosine similarity.
    scope_key: `str` (default: `None`)
        For each cell, cells with in scope_key are used for computing confidence.
        If `None`, use cell neighbors. Else, pick the cells in scope_key. Valid
        scope_key has to be in adata.obs.
    copy: `bool` (default: `False`)
        Return a copy instead of writing to adata.
    only_velocity_genes: `bool` (default: `False`)
        Only use velocity genes.
    only_high_spearman: `bool` (default: `False`)
        Only use high spearman.

    Returns
    -------
    velocity_length: `.obs`
        Length of the velocity vectors for each individual cell
    velocity_confidence: `.obs`
        Confidence for each cell
    """  # noqa E501

    adata = data.copy() if copy else data
    if vkey not in adata.layers.keys():
        raise ValueError("You need to run `tl.velocity` first.")
    if method not in ["corr", "cosine"]:
        raise ValueError("Method must be either 'corr' or 'cosine'.")
    if scope_key is not None and method == "corr":
        raise ValueError("Cannot use scope_key with method 'corr'.")

    V = np.array(adata.layers[vkey])

    # filter genes if needed
    tmp_filter = np.invert(np.isnan(np.sum(V, axis=0)))
    if only_velocity_genes and (f"{vkey}_genes" in adata.var.keys()):
        tmp_filter &= np.array(adata.var[f"{vkey}_genes"], dtype=bool)
    if only_high_spearman and ("spearmans_score" in adata.var.keys()):
        tmp_filter &= adata.var["spearmans_score"].values > 0.1
    V = V[:, tmp_filter]

    # zero mean, only need for correlation
    if method == "corr":
        V -= V.mean(1)[:, None]
    V_norm = l2_norm(V, axis=1)
    # normalize, only need for cosine similarity
    if method == "cosine":
        V /= V_norm[:, None]
    R = np.zeros(adata.n_obs)

    indices = (
        get_indices(dist=get_neighs(adata, "distances"))[0]
        if not scope_key
        else adata.obs[scope_key]  # the scope_key (e.g. cluster) of each cell
    )
    Vi_neighs_avg_cache = {}
    for i in range(adata.n_obs):
        if not scope_key:
            # use the neighbors of the cell
            Vi_neighs = V[indices[i]]
        else:
            # use the cells in scope_key
            if indices[i] not in Vi_neighs_avg_cache:
                Vi_neighs = V[indices == indices[i]]
                Vi_neighs_avg_cache[indices[i]] = Vi_neighs.mean(0)
        if method == "corr":
            Vi_neighs -= Vi_neighs.mean(1)[:, None]
            R[i] = np.mean(
                np.einsum("ij, j", Vi_neighs, V[i])
                / (l2_norm(Vi_neighs, axis=1) * V_norm[i])[None, :]
            )
        elif method == "cosine":
            # could compute mean first, because V has been normed
            Vi_neighs_avg = (
                Vi_neighs_avg_cache[indices[i]] if scope_key else Vi_neighs.mean(0)
            )
            R[i] = np.inner(V[i], Vi_neighs_avg)

    adata.obs[f"{vkey}_length"] = V_norm.round(2)
    adata.obs[f"{vkey}_confidence_{method}"] = R

    logg.hint(f"added '{vkey}_length' (adata.obs)")
    logg.hint(f"added '{vkey}_confidence_{method}' (adata.obs)")

    # if f"{vkey}_confidence_transition" not in adata.obs.keys():
    #     velocity_confidence_transition(adata, vkey)

    return adata if copy else None


def split_by_cluster():
    """split scores by cluster."""
    pass


# Code modified from function inner_cluster_coh from
# https://github.com/qiaochen/VeloAE/blob/main/veloproj/eval_util.py
def velocity_cosine(
    data,
    vkey="velocity",
    copy=False,
):
    pass


def inner_cluster_coh(adata, k_cluster, k_velocity, return_raw=False):
    """In-cluster Coherence Score.

    Args:
        adata (Anndata): Anndata object.
        k_cluster (str): key to the cluster column in adata.obs DataFrame.
        k_velocity (str): key to the velocity matrix in adata.obsm.
        return_raw (bool): return aggregated or raw scores.

    Returns:
        dict: all_scores indexed by cluster_edges
        or
        dict: mean scores indexed by cluster_edges
        float: averaged score over all cells.

    """
    clusters = np.unique(adata.obs[k_cluster])
    scores = {}
    all_scores = {}
    for cat in clusters:
        sel = adata.obs[k_cluster] == cat
        nbs = adata.uns["neighbors"]["indices"][sel]
        same_cat_nodes = map(lambda nodes: keep_type(adata, nodes, cat, k_cluster), nbs)
        velocities = adata.layers[k_velocity]
        cat_vels = velocities[sel]
        cat_score = [
            cosine_similarity(cat_vels[[ith]], velocities[nodes]).mean()
            for ith, nodes in enumerate(same_cat_nodes)
            if len(nodes) > 0
        ]
        all_scores[cat] = cat_score
        scores[cat] = np.mean(cat_score)

    if return_raw:
        return all_scores

    return scores, np.mean([sc for sc in scores.values()])


def keep_type(adata, nodes, target, k_cluster):
    """Select cells of targeted type

    Args:
        adata (Anndata): Anndata object.
        nodes (list): Indexes for cells
        target (str): Cluster name.
        k_cluster (str): Cluster key in adata.obs dataframe
    Returns:
        list: Selected cells.
    """
    return nodes[adata.obs[k_cluster][nodes].values == target]



from dgl import DGLGraph
import dgl.function as fn
import dgl
from .layers import GINLayer, ApplyNodeFunc, MLP
from dgl.nn.pytorch import GraphConv
import torch
import torch.nn as nn
import torch.nn.functional as F
from deepvelo.base import BaseModel


class MnistModel(BaseModel):
    def __init__(self, num_classes=10):
        super().__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, num_classes)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x, dim=1)


class VeloModel(BaseModel):
    def __init__(self, n_genes, layers=[256, 64]):
        super().__init__()
        self.fc1 = nn.Linear(2 * n_genes, layers[0])
        self.fc2 = nn.Linear(layers[0], layers[1])
        self.fc3 = nn.Linear(layers[1], 2 * n_genes)

    def forward(self, x_u, x_s):
        """
        right now it is jus mlp, and the complexity of the middle part does not make sense;
        Change it to the attention model and constrain the information flow
        """
        batch, n_gene = x_u.shape
        # x should be (batch, features=2*n_gene)
        x = torch.cat([x_u, x_s], dim=1)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))  # (batch, 64)
        x = self.fc3(x)  # (batch, genes*2)
        beta = x[:, 0:n_gene]  # (batch, genes)
        gamma = x[:, n_gene : 2 * n_gene]
        pred = beta * x_u + gamma * x_s
        return pred


class VeloTransformer(BaseModel):
    def __init__(self, n_genes, layers=[2, 64], emb_dim=32):
        super().__init__()
        self.emb = nn.Embedding(num_embeddings=2 * n_genes, embedding_dim=emb_dim)
        self.attn = nn.MultiheadAttention(embed_dim=emb_dim, num_heads=1)
        self.fc1 = nn.Linear(emb_dim, emb_dim)
        self.fc2 = nn.Linear(emb_dim, layers[0])
        self.fc3 = nn.Linear(layers[0] * 2 * n_genes, 2 * n_genes)

    def forward(self, x_u, x_s):
        """
        right now it is jus mlp, and the complexity of the middle part does not make sense;
        Change it to the attention model and constrain the information flow
        """
        batch, n_gene = x_u.shape
        # x should be (batch, features=2*n_gene)
        x = torch.cat([x_u, x_s], dim=1)
        x = x.unsqueeze(2)  # (batch, 2*genes, 1)
        x_emb = x * self.emb.weight  # (batch, 2*genes, emb_dim=32)
        x_emb = F.relu(self.fc1(x_emb))
        x_emb = x_emb.permute(1, 0, 2)  # (2*genes, batch, emb_dim)

        x, attn_output_weights = self.attn(
            x_emb, x_emb, x_emb
        )  # (2*genes, batch, emb_dim)
        # TODO: try an additional activation on the attention
        x = x.permute(1, 0, 2)  # (batch, 2*genes, emb_dim=32)
        x = F.relu(self.fc2(x)).reshape([batch, -1])  # (batch, 2*genes*d)
        x = self.fc3(x)  # (batch, genes*2)
        beta = x[:, 0:n_gene]  # (batch, genes)
        gamma = x[:, n_gene : 2 * n_gene]
        pred = beta * x_u + gamma * x_s
        return pred


class VeloGCN(BaseModel):
    def __init__(
        self,
        g,
        n_genes,
        layers=[64],
        fc_layer=False,
        pred_unspliced=False,
        dropout=0.2,
        activation=F.relu,
    ):
        super(VeloGCN, self).__init__()
        self.g = g  # the graph
        self.fc_layer = fc_layer
        self.pred_unspliced = pred_unspliced
        self.curent_kinetic_rates = {}

        self.layers = nn.ModuleList()
        # input layer
        self.layers.append(GraphConv(n_genes * 2, layers[0], activation=activation))
        # hidden layers
        for i in range(len(layers) - 1):
            self.layers.append(
                GraphConv(layers[i], layers[i + 1], activation=activation)
            )
        self.dropout = nn.Dropout(dropout)
        # output layer: which maps the embedded cell states tothe prediction of
        # kinetic rates for each gene. The intuition is we should know the cell's
        # developing pattern given the profile of current cell.
        if pred_unspliced:
            out_layer_dim = n_genes * 3
        else:
            out_layer_dim = n_genes * 2
        if fc_layer:
            self.layers.append(GraphConv(layers[-1], layers[-1]))
            self.fc = nn.Linear(layers[-1], out_layer_dim)
        else:
            self.layers.append(GraphConv(layers[-1], out_layer_dim))

    def forward(self, x_u, x_s):
        """
        right now it is jus mlp, and the complexity of the middle part does not make sense;
        Change it to the attention model and constrain the information flow
        """
        batch, n_gene = x_u.shape
        # h should be (batch, features=2*n_gene)
        h = torch.cat([x_u, x_s], dim=1)  # features
        for i, layer in enumerate(self.layers):
            h = layer(self.g, h)
            h = self.dropout(h)
        x = h

        # x = F.relu(self.fc1(x))
        # x = F.relu(self.fc2(x))  # (batch, 64)
        # x = self.fc3(x)  # (batch, genes*2)
        if self.fc_layer:
            x = self.fc(x)

        beta = x[:, 0:n_gene]  # (batch, genes)
        gamma = x[:, n_gene : 2 * n_gene]
        self.curent_kinetic_rates["beta"] = beta
        self.curent_kinetic_rates["gamma"] = gamma
        # if pred_unspliced, x is (batch, 3*n_genes)
        if self.pred_unspliced:
            pred = beta * x_u - gamma * x_s
            alpha = x[:, 2 * n_gene : 3 * n_gene]
            self.curent_kinetic_rates["alpha"] = alpha
            pred_u = alpha - beta * x_u
            pred = torch.cat([pred, pred_u], dim=1)
        else:
            pred = beta * x_u + gamma * x_s
        return pred

    def get_current_batch_kinetic_rates(self):
        return self.curent_kinetic_rates


# implement the model using GIN layers
class VeloGIN(BaseModel):
    def __init__(
        self,
        g,
        n_genes,
        layers=[64],
        n_mlp_layers=2,
        hidden_dim=64,
        res_connect=False,
        activation=F.relu,
    ):
        super(VeloGIN, self).__init__()
        self.g = g  # the graph
        self.layers = nn.ModuleList()
        # TODO(Haotian): test an embedding layer before the input
        # input layer
        mlp = MLP(n_mlp_layers, n_genes * 2, hidden_dim, layers[0])
        self.layers.append(
            GINLayer(
                ApplyNodeFunc(mlp),
                aggr_type="sum",
                dropout=0.0,
                batch_norm=True,
                residual=res_connect,
                init_eps=0,
                learn_eps=True,
            )
        )
        # hidden layers
        for i in range(len(layers) - 1):
            mlp = MLP(n_mlp_layers, layers[i], hidden_dim, layers[i + 1])
            self.layers.append(
                GINLayer(
                    ApplyNodeFunc(mlp),
                    aggr_type="sum",
                    dropout=0.0,
                    batch_norm=True,
                    residual=res_connect,
                    init_eps=0,
                    learn_eps=True,
                )
            )
        # output layer
        mlp = MLP(n_mlp_layers, layers[-1], hidden_dim, n_genes * 2)
        self.layers.append(
            GINLayer(
                ApplyNodeFunc(mlp),
                aggr_type="sum",
                dropout=0.0,
                batch_norm=True,
                residual=res_connect,
                init_eps=0,
                learn_eps=True,
            )
        )

    def forward(self, x_u, x_s):
        """
        right now it is jus mlp, and the complexity of the middle part does not make sense;
        Change it to the attention model and constrain the information flow
        """
        batch, n_gene = x_u.shape
        # h should be (batch, features=2*n_gene)
        h = torch.cat([x_u, x_s], dim=1)  # features
        for i, layer in enumerate(self.layers):
            h = layer(self.g, h)
        x = h

        # x = F.relu(self.fc1(x))
        # x = F.relu(self.fc2(x))  # (batch, 64)
        # x = self.fc3(x)  # (batch, genes*2)
        beta = x[:, 0:n_gene]  # (batch, genes)
        gamma = x[:, n_gene : 2 * n_gene]
        pred = beta * x_u + gamma * x_s
        return pred


# implement mini-batch using node-flow in DGL


class NodeUpdate(nn.Module):
    def __init__(self, in_feats, out_feats, activation=None, test=False, concat=False):
        super(NodeUpdate, self).__init__()
        self.linear = nn.Linear(in_feats, out_feats)
        self.activation = activation
        self.concat = concat
        self.test = test

    def forward(self, node):
        h = node.data["h"]
        if self.test:
            h = h * node.data["norm"]
        h = self.linear(h)
        # skip connection
        if self.concat:
            h = torch.cat((h, self.activation(h)), dim=1)
        elif self.activation:
            h = self.activation(h)
        return {"activation": h}


class VeloGCNNodeFlow(BaseModel):
    def __init__(self, n_genes, layers=[64], activation=F.relu, dropout=0, **kargs):
        super(VeloGCNNodeFlow, self).__init__()
        self.n_layers = len(layers)
        if dropout != 0:
            self.dropout = nn.Dropout(p=dropout)
        else:
            self.dropout = None
        self.layers = nn.ModuleList()
        # input layer
        self.layers.append(NodeUpdate(n_genes * 2, layers[0], activation, concat=False))
        # hidden layers
        for i in range(len(layers) - 1):
            self.layers.append(
                NodeUpdate(layers[i], layers[i + 1], activation, concat=False)
            )
        # output layer
        self.layers.append(NodeUpdate(layers[-1], n_genes * 2))

    def forward(self, nf):
        assert nf.layers._graph.num_layers == len(self.layers) + 1
        x_u = nf.layers[0].data["Ux_sz"]
        x_s = nf.layers[0].data["Sx_sz"]
        batch, n_gene = x_u.shape
        input_activation = torch.cat([x_u, x_s], dim=1)
        # gnn propagate and output x (batch, feature_size)
        x = self._graph_forward(nf, input_activation)

        beta = x[:, 0:n_gene]  # (batch, genes)
        gamma = x[:, n_gene : 2 * n_gene]
        x_u_out, x_s_out = nf.layers[-1].data["Ux_sz"], nf.layers[-1].data["Sx_sz"]
        pred = beta * x_u_out + gamma * x_s_out
        return pred

    def _graph_forward(self, nf, input_activation):
        nf.layers[0].data["activation"] = input_activation

        for i, layer in enumerate(self.layers):
            h = nf.layers[i].data.pop("activation")
            if self.dropout:
                h = self.dropout(h)
            nf.layers[i].data["h"] = h
            nf.block_compute(
                i,
                fn.copy_src(src="h", out="m"),
                lambda node: {"h": node.mailbox["m"].mean(dim=1)},
                layer,
            )

        h = nf.layers[-1].data.pop("activation")
        return h

    # Implement the Pytorch Geometric version here



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['deepvelo',
 'deepvelo.base',
 'deepvelo.data_loader',
 'deepvelo.logger',
 'deepvelo.model',
 'deepvelo.trainer',
 'deepvelo.utils']

package_data = \
{'': ['*']}

install_requires = \
['adjustText>=0.7.3,<0.8.0',
 'hnswlib>=0.6.2,<0.7.0',
 'igraph>=0.9.10,<0.10.0',
 'ipywidgets>=7.6.5,<8.0.0',
 'numpy>=1.21.1,<2.0.0',
 'scanpy>=1.8.2,<2.0.0',
 'scvelo>=0.2.4,<0.3.0',
 'seaborn>=0.11.2,<0.12.0',
 'torch>=1.2,<2.0',
 'tqdm>=4.62.3,<5.0.0',
 'umap-learn>=0.5.2,<0.6.0']

extras_require = \
{':extra != "gpu"': ['dgl>=0.4,!=0.8.0.post1'],
 'gpu': ['dgl-cu101>=0.4,!=0.8.0.post1']}

setup_kwargs = {
    'name': 'deepvelo',
    'version': '0.2.4',
    'description': 'Deep Velocity',
    'long_description': "# DeepVelo - A Deep Learning-based velocity estimation tool with cell-specific kinetic rates\n\nThis is the official implementation of the [DeepVelo](https://www.biorxiv.org/content/10.1101/2022.04.03.486877) method.\nDeepVelo employs cell-specific kinetic rates and provides more accurate RNA velocity estimates for complex differentiation and lineage decision events in heterogeneous scRNA-seq data. Please check out the paper for more details.\n\n![alt text](https://user-images.githubusercontent.com/11674033/171066682-a899377f-fae1-452a-8b67-8bc8c244b641.png)\n\n## Installation\n\n```bash\npip install deepvelo\n```\n\nThe `dgl` package is required, the cpu version is installed by default. Feel free to install the [dgl cuda](https://www.dgl.ai/pages/start.html) version for GPU acceleration.\n\n```bash\npip install dgl-cu101>=0.4.3 # an example for CUDA 10.1\n\n```\n\n### Install the development version\n\nWe use poetry to manage dependencies.\n\n```bash\npoetry install\n```\n\nThis will install the exact versions in the provided [poetry.lock](poetry.lock) file. If you want to install the latest version for all dependencies, use the following command.\n\n```bash\npoetry update\n```\n\n## Usage\n\nWe provide a number of notebooks in the [exmaples](examples) folder to help you get started. DeepVelo fullly integrates with [scanpy](https://scanpy.readthedocs.io/en/latest/) and [scVelo](https://scvelo.readthedocs.io/). The basic usage is as follows:\n\n```python\nimport deepvelo as dv\nimport scvelo as scv\n\nadata = ... # load your data in AnnData format\n\n# preprocess the data\nscv.pp.filter_and_normalize(adata, min_shared_counts=20, n_top_genes=2000)\nscv.pp.moments(adata, n_neighbors=30, n_pcs=30)\n\n# run DeepVelo using the default configs\ntrainer = dv.train(adata, dv.Constants.default_configs)\n# this will train the model and predict the velocity vectore. The result is stored in adata.layers['velocity']. You can use trainer.model to access the model.\n```\n",
    'author': 'subercui',
    'author_email': 'subercui@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bowang-lab/DeepVelo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<4.0',
}


setup(**setup_kwargs)



from scvelo import settings
from scvelo import logging as logg
from scvelo.preprocessing.moments import get_connectivities
from scvelo.tools.utils import make_unique_list, test_bimodality
from scvelo.tools.dynamical_model_utils import (
    BaseDynamics,
    linreg,
    convolve,
    tau_inv,
    unspliced,
)
from scvelo.tools.terminal_states import eigs

import numpy as np
import pandas as pd
import matplotlib.pyplot as pl
from matplotlib import rcParams
from scipy.optimize import minimize


class DynamicsRecovery(BaseDynamics):
    def __init__(self, adata=None, gene=None, load_pars=None, **kwargs):
        super(DynamicsRecovery, self).__init__(adata, gene, **kwargs)
        if load_pars and "fit_alpha" in adata.var.keys():
            self.load_pars(adata, gene)
        elif self.recoverable:
            self.initialize()

    def initialize(self):
        # set weights
        u, s, w, perc = self.u, self.s, self.weights, 98
        u_w, s_w, = (
            u[w],
            s[w],
        )

        # initialize scaling
        self.std_u, self.std_s = np.std(u_w), np.std(s_w)
        if self.std_u == 0 or self.std_s == 0:
            self.std_u = self.std_s = 1
        _scaling = self.fit_scaling
        scaling = self.std_u / self.std_s if isinstance(_scaling, bool) else _scaling
        u, u_w = u / scaling, u_w / scaling

        # initialize beta and gamma from extreme quantiles of s
        weights_s = s_w >= np.percentile(s_w, perc, axis=0)
        weights_u = u_w >= np.percentile(u_w, perc, axis=0)

        _prior = self.steady_state_prior
        weights_g = weights_s if _prior is None else weights_s | _prior[w]
        beta = 1
        gamma = linreg(convolve(u_w, weights_g), convolve(s_w, weights_g)) + 1e-6
        # 1e-6 to avoid beta = gamma
        # initialize gamma / beta * scaling clipped to adapt faster to extreme ratios
        if gamma < 0.05 / scaling:
            gamma *= 1.2
        elif gamma > 1.5 / scaling:
            gamma /= 1.2

        u_inf, s_inf = u_w[weights_u | weights_s].mean(), s_w[weights_s].mean()
        u0_, s0_ = u_inf, s_inf
        alpha = u_inf * beta
        # np.mean([s_inf * gamma, u_inf * beta])  # np.mean([s0_ * gamma, u0_ * beta])

        if self.init_vals is not None:  # just to test different EM start
            self.init_vals = np.array(self.init_vals)
            alpha, beta, gamma = np.array([alpha, beta, gamma]) * self.init_vals

        # initialize switching from u quantiles and alpha from s quantiles
        try:
            tstat_u, pval_u, means_u = test_bimodality(u_w, kde=True)
            tstat_s, pval_s, means_s = test_bimodality(s_w, kde=True)
        except:
            logg.warn("skipping bimodality check for", self.gene)
            tstat_u, tstat_s, pval_u, pval_s = 0, 0, 1, 1
            means_u, means_s = [0, 0], [0, 0]

        self.pval_steady = max(pval_u, pval_s)
        self.steady_u = means_u[1]
        self.steady_s = means_s[1]

        if self.pval_steady < 1e-3:
            u_inf = np.mean([u_inf, self.steady_u])
            alpha = gamma * s_inf
            beta = alpha / u_inf
            u0_, s0_ = u_inf, s_inf

        # alpha, beta, gamma = np.array([alpha, beta, gamma]) * scaling
        t_ = tau_inv(u0_, s0_, 0, 0, alpha, beta, gamma)

        # update object with initialized vars
        self.alpha, self.beta, self.gamma = alpha, beta, gamma
        self.scaling, self.alpha_ = scaling, 0
        self.u0_, self.s0_, self.t_ = u0_, s0_, t_
        self.pars = np.array([alpha, beta, gamma, self.t_, self.scaling])[:, None]

        # initialize time point assignment
        self.t, self.tau, self.o = self.get_time_assignment()
        self.loss = [self.get_loss()]

        self.initialize_scaling(sight=0.5)
        self.initialize_scaling(sight=0.1)

        self.steady_state_ratio = self.gamma / self.beta

        self.set_callbacks()

    def initialize_scaling(self, sight=0.5):  # fit scaling and update if improved
        z_vals = self.scaling + np.linspace(-1, 1, num=4) * self.scaling * sight
        for z in z_vals:
            self.update(scaling=z, beta=self.beta / self.scaling * z)

    def fit(self, assignment_mode=None):
        if self.max_iter > 0:
            # for comparison with exact time assignment
            if assignment_mode == "full_projection":
                self.assignment_mode = assignment_mode

            # pre-train with explicit time assignment
            self.fit_t_and_alpha()
            self.fit_scaling_()
            self.fit_rates()
            self.fit_t_()

            # actual EM (each iteration of simplex downhill is
            self.fit_t_and_rates()

            # train with optimal time assignment (oth. projection)
            self.assignment_mode = assignment_mode
            self.update(adjust_t_=False)
            self.fit_t_and_rates(refit_time=False)

        # self.update(adjust_t_=False)
        # self.t, self.tau, self.o = self.get_time_assignment()
        self.update()
        self.tau, self.tau_ = self.get_divergence(mode="tau")
        self.likelihood = self.get_likelihood(refit_time=False)
        self.varx = self.get_variance()

    def fit_alpha(self, sight=0.5, **kwargs):
        def mse(x):
            return self.get_mse(alpha=x[0], **kwargs)

        val = self.alpha
        vals = val + np.linspace(-1, 1, num=4) * val * sight
        for v in vals:
            self.update(alpha=val * v)
        res = minimize(mse, np.array([val]), **self.simplex_kwargs)
        self.update(alpha=res.x[0])

    def fit_beta(self, sight=0.5, **kwargs):
        def mse(x):
            return self.get_mse(beta=x[0], **kwargs)  # scaling=x[1])

        val = self.beta
        vals = val + np.linspace(-1, 1, num=4) * val * sight
        for v in vals:
            self.update(beta=val * v)
        res = minimize(mse, np.array([val]), **self.simplex_kwargs)
        self.update(beta=res.x[0])

    def fit_gamma(self, sight=0.5, **kwargs):
        def mse(x):
            return self.get_mse(gamma=x[0], **kwargs)

        val = self.gamma
        vals = val + np.linspace(-1, 1, num=4) * val * sight
        for v in vals:
            self.update(gamma=val * v)
        res = minimize(mse, np.array([val]), **self.simplex_kwargs)
        self.update(gamma=res.x[0])

    def fit_t_and_alpha(self, **kwargs):
        def mse(x):
            return self.get_mse(t_=x[0], alpha=x[1], **kwargs)

        alpha_vals = self.alpha + np.linspace(-1, 1, num=5) * self.alpha / 10
        for alpha in alpha_vals:
            self.update(alpha=alpha)
        x0, cb = np.array([self.t_, self.alpha]), self.cb_fit_t_and_alpha
        res = minimize(mse, x0, callback=cb, **self.simplex_kwargs)  # using Nelder-Mead
        self.update(t_=res.x[0], alpha=res.x[1])

    def fit_rates(self, **kwargs):
        def mse(x):
            return self.get_mse(alpha=x[0], gamma=x[1], **kwargs)

        x0, cb = np.array([self.alpha, self.gamma]), self.cb_fit_rates
        res = minimize(mse, x0, tol=1e-2, callback=cb, **self.simplex_kwargs)
        self.update(alpha=res.x[0], gamma=res.x[1])

    def fit_t_(self, **kwargs):
        def mse(x):
            return self.get_mse(t_=x[0], **kwargs)

        res = minimize(mse, self.t_, callback=self.cb_fit_t_, **self.simplex_kwargs)
        self.update(t_=res.x[0])

    def fit_rates_all(self, **kwargs):
        def mse(x):
            return self.get_mse(alpha=x[0], beta=x[1], gamma=x[2], **kwargs)

        x0, cb = np.array([self.alpha, self.beta, self.gamma]), self.cb_fit_rates_all
        res = minimize(mse, x0, tol=1e-2, callback=cb, **self.simplex_kwargs)
        self.update(alpha=res.x[0], beta=res.x[1], gamma=res.x[2])

    def fit_t_and_rates(self, **kwargs):
        def mse(x):
            return self.get_mse(t_=x[0], alpha=x[1], beta=x[2], gamma=x[3], **kwargs)

        x0 = np.array([self.t_, self.alpha, self.beta, self.gamma])
        cb = self.cb_fit_t_and_rates
        res = minimize(mse, x0, tol=1e-2, callback=cb, **self.simplex_kwargs)
        self.update(t_=res.x[0], alpha=res.x[1], beta=res.x[2], gamma=res.x[3])

    def fit_scaling_(self, **kwargs):
        def mse(x):
            return self.get_mse(t_=x[0], beta=x[1], scaling=x[2], **kwargs)

        x0, cb = np.array([self.t_, self.beta, self.scaling]), self.cb_fit_scaling_
        res = minimize(mse, x0, callback=cb, **self.simplex_kwargs)
        self.update(t_=res.x[0], beta=res.x[1], scaling=res.x[2])

    # Callback functions for the Optimizer
    def cb_fit_t_and_alpha(self, x):
        self.update(t_=x[0], alpha=x[1])

    def cb_fit_scaling_(self, x):
        self.update(t_=x[0], beta=x[1], scaling=x[2])

    def cb_fit_rates(self, x):
        self.update(alpha=x[0], gamma=x[1])

    def cb_fit_t_(self, x):
        self.update(t_=x[0])

    def cb_fit_t_and_rates(self, x):
        self.update(t_=x[0], alpha=x[1], beta=x[2], gamma=x[3])

    def cb_fit_rates_all(self, x):
        self.update(alpha=x[0], beta=x[1], gamma=x[2])

    def set_callbacks(self):
        # Overwrite callbacks
        if not self.high_pars_resolution:
            self.cb_fit_t_and_alpha = None
            self.cb_fit_scaling_ = None
            self.cb_fit_rates = None
            self.cb_fit_t_ = None
            self.cb_fit_t_and_rates = None
            self.cb_fit_rates_all = None

    def update(
        self,
        t=None,
        t_=None,
        alpha=None,
        beta=None,
        gamma=None,
        scaling=None,
        u0_=None,
        s0_=None,
        adjust_t_=True,
    ):
        loss_prev = self.loss[-1] if len(self.loss) > 0 else 1e6

        _vars = self.get_vars(alpha, beta, gamma, scaling, t_, u0_, s0_)
        _time = self.get_time_assignment(alpha, beta, gamma, scaling, t_, u0_, s0_, t)
        alpha, beta, gamma, scaling, t_ = _vars
        t, tau, o = _time
        loss = self.get_loss(t, t_, alpha, beta, gamma, scaling)
        perform_update = loss < loss_prev

        on = self.o == 1
        if adjust_t_ and np.any(on):
            if not perform_update:
                alpha, beta, gamma, scaling, t_ = self.get_vars()
                t, tau, o = self.get_time_assignment()
                loss = self.get_loss()

            alt_t_ = t[on].max()
            if 0 < alt_t_ < t_:
                # alt_u0_, alt_s0_ = mRNA(alt_t_, 0, 0, alpha, beta, gamma)
                alt_t_ += np.max(t) / len(t) * np.sum(t == t_)
                # np.sum((self.u / self.scaling >= alt_u0_) | (self.s >= alt_s0_))
                _time = self.get_time_assignment(alpha, beta, gamma, scaling, alt_t_)
                alt_t, alt_tau, alt_o = _time
                alt_loss = self.get_loss(alt_t, alt_t_, alpha, beta, gamma, scaling)
                ut_cur = unspliced(t_, 0, alpha, beta)
                ut_alt = unspliced(alt_t_, 0, alpha, beta)

                min_loss = np.min([loss, loss_prev])
                if alt_loss * 0.99 <= min_loss or ut_cur * 0.99 < ut_alt:
                    t, tau, o, t_, loss = alt_t, alt_tau, alt_o, alt_t_, alt_loss
                    perform_update = True

            if False:
                steady_states = t == t_
                if perform_update and np.any(steady_states):
                    t_ += t.max() / len(t) * np.sum(steady_states)
                    _time = self.get_time_assignment(alpha, beta, gamma, scaling, t_)
                    t, tau, o = _time
                    loss = self.get_loss(t, t_, alpha, beta, gamma, scaling)

        if perform_update:
            if scaling is not None:
                self.steady_u *= self.scaling / scaling
                self.u0_ *= self.scaling / scaling
            if u0_ is not None:
                self.u0_ = u0_
            if s0_ is not None:
                self.s0_ = s0_

            self.t, self.tau, self.o = t, tau, o
            self.alpha, self.beta, self.gamma = alpha, beta, gamma
            self.scaling, self.t_ = scaling, t_
            new_pars = np.array([alpha, beta, gamma, t_, scaling])[:, None]
            self.pars = np.c_[self.pars, new_pars]
            self.loss.append(loss)

        return perform_update


default_pars_names = ["alpha", "beta", "gamma", "t_", "scaling", "std_u", "std_s"]
default_pars_names += ["likelihood", "u0", "s0", "pval_steady"]
default_pars_names += ["steady_u", "steady_s", "variance"]


def read_pars(adata, pars_names=None, key="fit"):
    pars = []
    for name in default_pars_names if pars_names is None else pars_names:
        pkey = f"{key}_{name}"
        par = np.zeros(adata.n_vars) * np.nan
        if pkey in adata.var.keys():
            par = adata.var[pkey].values
        pars.append(par)
    return pars


def write_pars(adata, pars, pars_names=None, add_key="fit"):
    for i, name in enumerate(default_pars_names if pars_names is None else pars_names):
        adata.var[f"{add_key}_{name}"] = pars[i]


def recover_dynamics(
    data,
    var_names="velocity_genes",
    n_top_genes=None,
    max_iter=10,
    assignment_mode="projection",
    t_max=None,
    fit_time=True,
    fit_scaling=True,
    fit_steady_states=True,
    fit_connected_states=None,
    fit_basal_transcription=None,
    use_raw=False,
    load_pars=None,
    return_model=None,
    plot_results=False,
    steady_state_prior=None,
    add_key="fit",
    copy=False,
    **kwargs,
):
    """Recovers the full splicing kinetics of specified genes.

    The model infers transcription rates, splicing rates, degradation rates,
    as well as cell-specific latent time and transcriptional states,
    estimated iteratively by expectation-maximization.

    .. image:: https://user-images.githubusercontent.com/31883718/69636459-ef862800-1056-11ea-8803-0a787ede5ce9.png

    Arguments
    ---------
    data: :class:`~anndata.AnnData`
        Annotated data matrix.
    var_names: `str`,  list of `str` (default: `'velocity_genes`)
        Names of variables/genes to use for the fitting.
    n_top_genes: `int` or `None` (default: `None`)
        Number of top velocity genes to use for the dynamical model.
    max_iter:`int` (default: `10`)
        Maximal iterations in the EM-Algorithm.
    assignment_mode: `str` (default: `projection`)
        Determined how times are assigned to observations.
        If `projection`, observations are projected onto the model trajectory.
        Else uses an inverse approximating formula.
    t_max: `float`, `False` or `None` (default: `None`)
        Total range for time assignments.
    fit_scaling: `bool` or `float` or `None` (default: `True`)
        Whether to fit scaling between unspliced and spliced.
    fit_time: `bool` or `float` or `None` (default: `True`)
        Whether to fit time or keep initially given time fixed.
    fit_steady_states: `bool` or `None` (default: `True`)
        Whether to explicitly model and fit steady states (next to induction/repression)
    fit_connected_states: `bool` or `None` (default: `None`)
        Restricts fitting to neighbors given by connectivities.
    fit_basal_transcription: `bool` or `None` (default: `None`)
        Enables model to incorporate basal transcriptions.
    use_raw: `bool` or `None` (default: `None`)
        if True, use .layers['sliced'], else use moments from .layers['Ms']
    load_pars: `bool` or `None` (default: `None`)
        Load parameters from past fits.
    return_model: `bool` or `None` (default: `True`)
        Whether to return the model as :DynamicsRecovery: object.
    plot_results: `bool` or `None` (default: `False`)
        Plot results after parameter inference.
    steady_state_prior: list of `bool` or `None` (default: `None`)
        Mask for indices used for steady state regression.
    add_key: `str` (default: `'fit'`)
        Key to add to parameter names, e.g. 'fit_t' for fitted time.
    copy: `bool` (default: `False`)
        Return a copy instead of writing to `adata`.

    Returns
    -------
    Returns or updates `adata`
    """
    adata = data.copy() if copy else data
    logg.info("recovering dynamics", r=True)

    if len(set(adata.var_names)) != len(adata.var_names):
        logg.warn("Duplicate var_names found. Making them unique.")
        adata.var_names_make_unique()

    if "Ms" not in adata.layers.keys() or "Mu" not in adata.layers.keys():
        use_raw = True
    if fit_connected_states is None:
        fit_connected_states = not use_raw

    adata.uns["recover_dynamics"] = {
        "fit_connected_states": fit_connected_states,
        "fit_basal_transcription": fit_basal_transcription,
        "use_raw": use_raw,
    }

    if isinstance(var_names, str) and var_names not in adata.var_names:
        if var_names in adata.var.keys():
            var_names = adata.var_names[adata.var[var_names].values]
        elif use_raw or var_names == "all":
            var_names = adata.var_names
        elif "_genes" in var_names:
            from scvelo.tools.velocity import Velocity

            velo = Velocity(adata, use_raw=use_raw)
            velo.compute_deterministic(perc=[5, 95])
            var_names = adata.var_names[velo._velocity_genes]
            adata.var["fit_r2"] = velo._r2
        else:
            raise ValueError("Variable name not found in var keys.")
    if not isinstance(var_names, str):
        var_names = list(np.ravel(var_names))

    var_names = make_unique_list(var_names, allow_array=True)
    var_names = np.array([name for name in var_names if name in adata.var_names])
    if len(var_names) == 0:
        raise ValueError("Variable name not found in var keys.")
    if n_top_genes is not None and len(var_names) > n_top_genes:
        X = adata[:, var_names].layers[("spliced" if use_raw else "Ms")]
        var_names = var_names[np.argsort(np.sum(X, 0))[::-1][:n_top_genes]]
    if return_model is None:
        return_model = len(var_names) < 5

    pars = read_pars(adata)
    alpha, beta, gamma, t_, scaling, std_u, std_s, likelihood = pars[:8]
    u0, s0, pval, steady_u, steady_s, varx = pars[8:]
    # likelihood[np.isnan(likelihood)] = 0
    idx, L, P = [], [], []
    T = np.zeros(adata.shape) * np.nan
    Tau = np.zeros(adata.shape) * np.nan
    Tau_ = np.zeros(adata.shape) * np.nan
    if "fit_t" in adata.layers.keys():
        T = adata.layers["fit_t"]
    if "fit_tau" in adata.layers.keys():
        Tau = adata.layers["fit_tau"]
    if "fit_tau_" in adata.layers.keys():
        Tau_ = adata.layers["fit_tau_"]

    conn = get_connectivities(adata) if fit_connected_states else None
    progress = logg.ProgressReporter(len(var_names))
    for i, gene in enumerate(var_names):
        dm = DynamicsRecovery(
            adata,
            gene,
            use_raw=use_raw,
            load_pars=load_pars,
            max_iter=max_iter,
            fit_time=fit_time,
            fit_steady_states=fit_steady_states,
            fit_connected_states=conn,
            fit_scaling=fit_scaling,
            fit_basal_transcription=fit_basal_transcription,
            steady_state_prior=steady_state_prior,
            **kwargs,
        )
        if dm.recoverable:
            dm.fit(assignment_mode=assignment_mode)

            ix = np.where(adata.var_names == gene)[0][0]
            idx.append(ix)

            T[:, ix], Tau[:, ix], Tau_[:, ix] = dm.t, dm.tau, dm.tau_
            alpha[ix], beta[ix], gamma[ix], t_[ix], scaling[ix] = dm.pars[:, -1]
            u0[ix], s0[ix], pval[ix] = dm.u0, dm.s0, dm.pval_steady
            steady_u[ix], steady_s[ix] = dm.steady_u, dm.steady_s
            beta[ix] /= scaling[ix]
            steady_u[ix] *= scaling[ix]

            std_u[ix], std_s[ix] = dm.std_u, dm.std_s
            likelihood[ix], varx[ix] = dm.likelihood, dm.varx
            L.append(dm.loss)
            if plot_results and i < 4:
                P.append(np.array(dm.pars))

            progress.update()
        else:
            logg.warn(dm.gene, "not recoverable due to insufficient samples.")
    progress.finish()

    _pars = [
        alpha,
        beta,
        gamma,
        t_,
        scaling,
        std_u,
        std_s,
        likelihood,
        u0,
        s0,
        pval,
        steady_u,
        steady_s,
        varx,
    ]
    write_pars(adata, _pars)
    if "fit_t" in adata.layers.keys():
        adata.layers["fit_t"][:, idx] = (
            T[:, idx] if conn is None else conn.dot(T[:, idx])
        )
    else:
        adata.layers["fit_t"] = T if conn is None else conn.dot(T)
    adata.layers["fit_tau"] = Tau
    adata.layers["fit_tau_"] = Tau_

    if L:  # is False if only one invalid / irrecoverable gene was given in var_names
        cur_len = adata.varm["loss"].shape[1] if "loss" in adata.varm.keys() else 2
        max_len = max(np.max([len(l) for l in L]), cur_len) if L else cur_len
        loss = np.ones((adata.n_vars, max_len)) * np.nan

        if "loss" in adata.varm.keys():
            loss[:, :cur_len] = adata.varm["loss"]

        loss[idx] = np.vstack(
            [np.concatenate([l, np.ones(max_len - len(l)) * np.nan]) for l in L]
        )
        adata.varm["loss"] = loss

    if t_max is not False:
        dm = align_dynamics(adata, t_max=t_max, dm=dm, idx=idx)

    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
    logg.hint(
        "added \n"
        f"    '{add_key}_pars', "
        f"fitted parameters for splicing dynamics (adata.var)"
    )

    if plot_results:  # Plot Parameter Stats
        n_rows, n_cols = len(var_names[:4]), 6
        figsize = [2 * n_cols, 1.5 * n_rows]  # rcParams['figure.figsize']
        fontsize = rcParams["font.size"]
        fig, axes = pl.subplots(nrows=n_rows, ncols=6, figsize=figsize)
        pl.subplots_adjust(wspace=0.7, hspace=0.5)
        for i, gene in enumerate(var_names[:4]):
            if t_max is not False:
                mi = dm.m[i]
                P[i] *= np.array([1 / mi, 1 / mi, 1 / mi, mi, 1])[:, None]
            ax = axes[i] if n_rows > 1 else axes
            for j, pij in enumerate(P[i]):
                ax[j].plot(pij)
            ax[len(P[i])].plot(L[i])
            if i == 0:
                pars_names = ["alpha", "beta", "gamma", "t_", "scaling", "loss"]
                for j, name in enumerate(pars_names):
                    ax[j].set_title(name, fontsize=fontsize)

    if return_model:
        logg.info("\noutputs model fit of gene:", dm.gene)

    return dm if return_model else adata if copy else None


def align_dynamics(
    data, t_max=None, dm=None, idx=None, mode=None, remove_outliers=None, copy=False
):
    """Align dynamics to a common set of parameters

    Arguments
    ---------
    data: :class:`~anndata.AnnData`
        Annotated data matrix.
    t_max: `float`, `False` or `None` (default: `None`)
        Total range for time assignments.
    dm: :class:`~DynamicsRecovery`
        DynamicsRecovery object to perform alignment on.
    idx: list of `bool` or `None` (default: `None`)
        Mask for indices used for alignment.
    mode: `str` or None (default: `'align_total_time`)
        What to align. Takes the following arguments:
        common_splicing_rate, common_scaling, align_increments, align_total_time
    remove_outliers: `bool` or `None` (default: `None`)
        Whether to remove outliers.
    copy: `bool` (default: `False`)
        Return a copy instead of writing to `adata`.
     Returns
    -------
    Returns or updates `adata` with the attributes
    alpha, beta, gamma, t_, alignment_scaling: `.var`
        aligned parameters
    fit_t, fit_tau, fit_tau_: `.layer`
        aligned time
    """

    adata = data.copy() if copy else data
    pars_names = ["alpha", "beta", "gamma", "t_", "scaling", "alignment_scaling"]
    alpha, beta, gamma, t_, scaling, mz = read_pars(adata, pars_names=pars_names)
    T = np.zeros(adata.shape) * np.nan
    Tau = np.zeros(adata.shape) * np.nan
    Tau_ = np.zeros(adata.shape) * np.nan
    if "fit_t" in adata.layers.keys():
        T = adata.layers["fit_t"]
    if "fit_tau" in adata.layers.keys():
        Tau = adata.layers["fit_tau"]
    if "fit_tau_" in adata.layers.keys():
        Tau_ = adata.layers["fit_tau_"]
    idx = ~np.isnan(np.sum(T, axis=0)) if idx is None else idx
    if "fit_alignment_scaling" not in adata.var.keys():
        mz = np.ones(adata.n_vars)
    if mode is None:
        mode = "align_total_time"

    m = np.ones(adata.n_vars)
    mz_prev = np.array(mz)

    if dm is not None:  # newly fitted
        mz[idx] = 1

    if mode == "align_total_time" and t_max is not False:
        T_max = np.max(T[:, idx] * (T[:, idx] < t_[idx]), axis=0)  # transient 'on'
        T_max += np.max((T[:, idx] - t_[idx]) * (T[:, idx] > t_[idx]), axis=0)  # 'off'

        denom = 1 - np.sum((T[:, idx] == t_[idx]) | (T[:, idx] == 0), axis=0) / len(T)
        denom += denom == 0

        T_max = T_max / denom
        T_max += T_max == 0

        t_max = 20 if t_max is None else t_max
        m[idx] = t_max / T_max
        mz *= m

    else:
        m = 1 / mz
        mz = np.ones(adata.n_vars)

    if remove_outliers:
        mu, std = np.nanmean(mz), np.nanstd(mz)
        mz = np.clip(mz, mu - 3 * std, mu + 3 * std)
        m = mz / mz_prev

    if idx is None:
        alpha, beta, gamma = alpha / m, beta / m, gamma / m
        T, t_, Tau, Tau_ = T * m, t_ * m, Tau * m, Tau_ * m
    else:
        m_ = m[idx]
        alpha[idx] = alpha[idx] / m_
        beta[idx] = beta[idx] / m_
        gamma[idx] = gamma[idx] / m_
        T[:, idx], t_[idx] = T[:, idx] * m_, t_[idx] * m_
        Tau[:, idx], Tau_[:, idx] = Tau[:, idx] * m_, Tau_[:, idx] * m_

    mz[mz == 1] = np.nan
    pars_names = ["alpha", "beta", "gamma", "t_", "alignment_scaling"]
    write_pars(adata, [alpha, beta, gamma, t_, mz], pars_names=pars_names)
    adata.layers["fit_t"] = T
    adata.layers["fit_tau"] = Tau
    adata.layers["fit_tau_"] = Tau_

    if dm is not None and dm.recoverable:
        dm.m = m[idx]
        dm.alpha, dm.beta, dm.gamma, dm.pars[:3] = (
            np.array([dm.alpha, dm.beta, dm.gamma, dm.pars[:3]]) / dm.m[-1]
        )
        dm.t, dm.tau, dm.t_, dm.pars[4] = (
            np.array([dm.t, dm.tau, dm.t_, dm.pars[4]]) * dm.m[-1]
        )

    return adata if copy else dm


def latent_time(
    data,
    vkey="velocity",
    min_likelihood=0.1,
    min_confidence=0.75,
    min_corr_diffusion=None,
    weight_diffusion=None,
    root_key=None,
    end_key=None,
    t_max=None,
    use_fit_t=False,
    method="dpt",
    copy=False,
):
    """Computes a gene-shared latent time.

    Gene-specific latent timepoints obtained from the dynamical model are coupled to a
    universal gene-shared latent time, which represents the cell’s internal clock and
    is based only on its transcriptional dynamics.

    .. image:: https://user-images.githubusercontent.com/31883718/69636500-03318e80-1057-11ea-9e14-ae9f907711cc.png

    Arguments
    ---------
    data: :class:`~anndata.AnnData`
        Annotated data matrix
    vkey: `str` (default: `'velocity'`)
        Name of velocity estimates to be used.
    min_likelihood: `float` between `0` and `1` or `None` (default: `.1`)
        Minimal likelihood fitness for genes to be included to the weighting.
    min_confidence: `float` between `0` and `1` (default: `.75`)
        Parameter for local coherence selection.
    min_corr_diffusion: `float` between `0` and `1` or `None` (default: `None`)
        Only select genes that correlate with velocity pseudotime obtained
        from diffusion random walk on velocity graph.
    weight_diffusion: `float` or `None` (default: `None`)
        Weight applied to couple latent time with diffusion-based velocity pseudotime.
    root_key: `str` or `None` (default: `'root_cells'`)
        Key (.uns, .obs) of root cell to be used.
        If not set, it obtains root cells from velocity-inferred transition matrix.
    end_key: `str` or `None` (default: `None`)
        Key (.obs) of end points to be used.
    use_fit_t: `bool` (default: `false`)
        Use fit_t in the dynamical model to calculate the latent time.
    method: `str` or `None` (default: `'dpt1'`)
        The method to compute the relative time.
    t_max: `float` or `None` (default: `None`)
        Overall duration of differentiation process.
        If not set, a overall transcriptional timescale of 20 hours is used as prior.
    copy: `bool` (default: `False`)
        Return a copy instead of writing to `adata`.
     Returns
    -------
    Returns or updates `adata` with the attributes
    latent_time: `.obs`
        latent time from learned dynamics for each cell
    """
    adata = data.copy() if copy else data

    from scvelo.tools.utils import vcorrcoef, scale
    from scvelo.tools.dynamical_model_utils import root_time, compute_shared_time
    from scvelo.tools.terminal_states import terminal_states
    from scvelo.tools.velocity_graph import velocity_graph
    from scvelo.tools.velocity_pseudotime import velocity_pseudotime

    if use_fit_t and "fit_t" not in adata.layers.keys():
        raise ValueError("you need to run `tl.recover_dynamics` first.")

    # NOTE(Haotian): first need the velocity graph
    if f"{vkey}_graph" not in adata.uns.keys():
        velocity_graph(adata, approx=True)

    if root_key is None:
        terminal_keys = ["root_cells", "starting_cells", "root_states_probs"]
        keys = [key for key in terminal_keys if key in adata.obs.keys()]
        if len(keys) > 0:
            root_key = keys[0]
    if root_key not in adata.uns.keys() and root_key not in adata.obs.keys():
        root_key = "root_cells"
    # NOTE(Haotian): computing the root cells
    if root_key not in adata.obs.keys():
        terminal_states(adata, vkey=vkey)

    if use_fit_t:
        t = np.array(adata.layers["fit_t"])
        idx_valid = ~np.isnan(
            t.sum(0)
        )  # this is the valid idx on the gene axis, each has to have valid time t across all cells
        if min_likelihood is not None:  # filter out more unreliable ts
            likelihood = adata.var["fit_likelihood"].values
            idx_valid &= np.array(likelihood >= min_likelihood, dtype=bool)
        t = t[:, idx_valid]
        t_sum = np.sum(t, 1)
    conn = get_connectivities(adata)

    if root_key not in adata.uns.keys():
        roots = np.arange(adata.shape[0])  # instead of np.argsort(t_sum)
        idx_roots = np.array(adata.obs[root_key][roots])
        idx_roots[pd.isnull(idx_roots)] = 0
        if np.any([isinstance(ix, str) for ix in idx_roots]):
            idx_roots = np.array([isinstance(ix, str) for ix in idx_roots], dtype=int)
        idx_roots = idx_roots.astype(np.float) > 1 - 1e-3
        if np.sum(idx_roots) > 0:
            roots = roots[
                idx_roots
            ]  # NOTE(Haotian): 1. Here is the actual root_cell index. 2. I don't think the argsort(8 lines before) is needed, you just need the index.
        else:
            logg.warn(
                "No root cells detected. Consider specifying "
                "root cells to improve latent time prediction."
            )
    else:
        roots = [adata.uns[root_key]]
        root_key = f"root cell {adata.uns[root_key]}"

    if end_key in adata.obs.keys():
        fates = np.argsort(t_sum)[::-1]
        idx_fates = np.array(adata.obs[end_key][fates])
        idx_fates[pd.isnull(idx_fates)] = 0
        if np.any([isinstance(ix, str) for ix in idx_fates]):
            idx_fates = np.array([isinstance(ix, str) for ix in idx_fates], dtype=int)
        idx_fates = idx_fates.astype(np.float) > 1 - 1e-3
        if np.sum(idx_fates) > 0:
            fates = fates[idx_fates]
    else:
        fates = [None]

    logg.info(
        f"computing latent time using {root_key}"
        f"{', ' + end_key if end_key in adata.obs.keys() else ''} as prior",
        r=True,
    )

    # NOTE(Haotian): actual pseudotime call
    VPT = velocity_pseudotime(
        adata, vkey, root_key=roots[0], end_key=fates[0], return_model=True
    )
    vpt = VPT.pseudotime

    if use_fit_t:
        if min_corr_diffusion is not None:
            corr = vcorrcoef(t.T, vpt)
            t = t[:, np.array(corr >= min_corr_diffusion, dtype=bool)]

        if root_key in adata.uns.keys():
            root = adata.uns[root_key]
            t, t_ = root_time(t, root=root)
            latent_time = compute_shared_time(t)
        else:
            roots = roots[:4]  # NOTE(Haotian): you only use the top 4 cells?
            latent_time = np.ones(shape=(len(roots), adata.n_obs))
            for i, root in enumerate(roots):
                t, t_ = root_time(t, root=root)
                latent_time[i] = compute_shared_time(
                    t
                )  # NOTE(Haotian): for each one cell, this is super ad-hoc, just select those ts look better for 2 select rows.
            latent_time = scale(np.mean(latent_time, axis=0))

        if fates[0] is not None:
            fates = fates[:4]
            latent_time_ = np.ones(shape=(len(fates), adata.n_obs))
            for i, fate in enumerate(fates):
                t, t_ = root_time(t, root=fate)
                latent_time_[i] = 1 - compute_shared_time(t)
            latent_time = scale(
                latent_time + 0.2 * scale(np.mean(latent_time_, axis=0))
            )
    elif method == "eig":
        # using the eigenvector method
        T = data.uns[f"{vkey}_graph"] - data.uns[f"{vkey}_graph_neg"]
        # T += T.T
        eigvecs = eigs(T, eps=1e-3, random_state=0)[1]
        pt = T.dot(eigvecs).sum(1)
        # pt = scale(np.clip(pt, 0, np.percentile(pt, 98)))
        latent_time = scale(pt)
    else:
        latent_time = scale(vpt)

    tl = latent_time
    tc = conn.dot(latent_time)

    z = tl.dot(tc) / tc.dot(
        tc
    )  # NOTE(Haotian): what is the purpose of these steps? what is going on here? if z is closed to 1. basically, it is the eigenvector.
    tl_conf = (1 - np.abs(tl / np.max(tl) - tc * z / np.max(tl))) ** 2
    idx_low_confidence = tl_conf < min_confidence

    if weight_diffusion is not None:
        w = weight_diffusion
        latent_time = (1 - w) * latent_time + w * vpt
        latent_time[idx_low_confidence] = vpt[idx_low_confidence]
    else:
        conn_new = conn.copy()
        conn_new[:, idx_low_confidence] = 0
        conn_new.eliminate_zeros()
        latent_time = conn_new.dot(latent_time)

    latent_time = scale(latent_time)
    if t_max is not None:
        latent_time *= t_max

    adata.obs["latent_time"] = latent_time

    logg.info("    finished", time=True, end=" " if settings.verbosity > 2 else "\n")
    logg.hint("added \n" "    'latent_time', shared time (adata.obs)")
    return adata if copy else None


recover_latent_time = latent_time



import time
import torch
from abc import abstractmethod
from numpy import inf
from deepvelo.logger import TensorboardWriter


class BaseTrainer:
    """
    Base class for all trainers
    """

    def __init__(self, model, criterion, metric_ftns, optimizer, config):
        self.config = config
        self.logger = config.get_logger("trainer", config["trainer"]["verbosity"])

        # setup GPU device if available, move model into configured device
        self.device, device_ids = self._prepare_device(config["n_gpu"])
        self.model = model.to(self.device)
        if len(device_ids) > 1:
            self.model = torch.nn.DataParallel(model, device_ids=device_ids)

        self.criterion = criterion
        self.metric_ftns = metric_ftns
        self.optimizer = optimizer

        cfg_trainer = config["trainer"]
        self.epochs = cfg_trainer["epochs"]
        self.save_period = cfg_trainer["save_period"]
        self.monitor = cfg_trainer.get("monitor", "off")

        # configuration to monitor model performance and save best
        if self.monitor == "off":
            self.mnt_mode = "off"
            self.mnt_best = 0
        else:
            self.mnt_mode, self.mnt_metric = self.monitor.split()
            assert self.mnt_mode in ["min", "max"]

            self.mnt_best = inf if self.mnt_mode == "min" else -inf
            self.early_stop = cfg_trainer.get("early_stop", inf)

        self.start_epoch = 1

        self.checkpoint_dir = config.save_dir

        # setup visualization writer instance
        self.writer = TensorboardWriter(
            config.log_dir, self.logger, cfg_trainer["tensorboard"]
        )

        if config.resume is not None:
            self._resume_checkpoint(config.resume)

    @abstractmethod
    def _train_epoch(self, epoch):
        """
        Training logic for an epoch

        :param epoch: Current epoch number
        """
        raise NotImplementedError

    def train(self, callback=None, callback_freq=1):
        """
        Full training logic
        """
        not_improved_count = 0
        tik = time.time()
        if "mle" in self.config["loss"]["type"]:
            if self.config["arch"]["args"]["pred_unspliced"]:
                self.candidate_states = torch.cat(
                    [
                        self.data_loader.dataset.Sx_sz,
                        self.data_loader.dataset.Ux_sz,
                    ],
                    dim=1,
                ).to(self.device)
            else:
                self.candidate_states = self.data_loader.dataset.Sx_sz.to(self.device)
        for epoch in range(self.start_epoch, self.epochs + 1):
            result = self._train_epoch(epoch)

            # save logged informations into log dict
            log = {"epoch": epoch, "time:": time.time() - tik}
            log.update(result)
            tik = time.time()

            # print logged informations to the screen
            for key, value in log.items():
                self.logger.info("    {:15s}: {}".format(str(key), value))

            if callback is not None:
                if epoch % callback_freq == 0:
                    callback(epoch)

            # evaluate model performance according to configured metric, save best checkpoint as model_best
            best = False
            if self.mnt_mode != "off":
                try:
                    # check whether model performance improved or not, according to specified metric(mnt_metric)
                    improved = (
                        self.mnt_mode == "min" and log[self.mnt_metric] <= self.mnt_best
                    ) or (
                        self.mnt_mode == "max" and log[self.mnt_metric] >= self.mnt_best
                    )
                except KeyError:
                    self.logger.warning(
                        "Warning: Metric '{}' is not found. "
                        "Model performance monitoring is disabled.".format(
                            self.mnt_metric
                        )
                    )
                    self.mnt_mode = "off"
                    improved = False

                if improved:
                    self.mnt_best = log[self.mnt_metric]
                    not_improved_count = 0
                    best = True
                else:
                    not_improved_count += 1

                if not_improved_count > self.early_stop:
                    self.logger.info(
                        "Validation performance didn't improve for {} epochs. "
                        "Training stops.".format(self.early_stop)
                    )
                    break

            if epoch % self.save_period == 0:
                self._save_checkpoint(epoch, save_best=best)

    def train_with_epoch_callback(self, callback, freq):
        self.train(callback, freq)

    def _prepare_device(self, n_gpu_use):
        """
        setup GPU device if available, move model into configured device
        """
        n_gpu = torch.cuda.device_count()
        if n_gpu_use > 0 and n_gpu == 0:
            self.logger.warning(
                "Warning: There's no GPU available on this machine,"
                "training will be performed on CPU."
            )
            n_gpu_use = 0
        if n_gpu_use > n_gpu:
            self.logger.warning(
                "Warning: The number of GPU's configured to use is {}, but only {} are available "
                "on this machine.".format(n_gpu_use, n_gpu)
            )
            n_gpu_use = n_gpu
        device = torch.device("cuda:0" if n_gpu_use > 0 else "cpu")
        list_ids = list(range(n_gpu_use))
        return device, list_ids

    def _save_checkpoint(self, epoch, save_best=False):
        """
        Saving checkpoints

        :param epoch: current epoch number
        :param log: logging information of the epoch
        :param save_best: if True, rename the saved checkpoint to 'model_best.pth'
        """
        arch = type(self.model).__name__
        state = {
            "arch": arch,
            "epoch": epoch,
            "state_dict": self.model.state_dict(),
            "optimizer": self.optimizer.state_dict(),
            "monitor_best": self.mnt_best,
            "config": self.config,
        }
        filename = str(self.checkpoint_dir / "checkpoint-epoch{}.pth".format(epoch))
        torch.save(state, filename)
        self.logger.info("Saving checkpoint: {} ...".format(filename))
        if save_best:
            best_path = str(self.checkpoint_dir / "model_best.pth")
            torch.save(state, best_path)
            self.logger.info("Saving current best: model_best.pth ...")

    def _resume_checkpoint(self, resume_path):
        """
        Resume from saved checkpoints

        :param resume_path: Checkpoint path to be resumed
        """
        resume_path = str(resume_path)
        self.logger.info("Loading checkpoint: {} ...".format(resume_path))
        checkpoint = torch.load(resume_path)
        self.start_epoch = checkpoint["epoch"] + 1
        self.mnt_best = checkpoint["monitor_best"]

        # load architecture params from checkpoint.
        if checkpoint["config"]["arch"] != self.config["arch"]:
            self.logger.warning(
                "Warning: Architecture configuration given in config file is different from that of "
                "checkpoint. This may yield an exception while state_dict is being loaded."
            )
        self.model.load_state_dict(checkpoint["state_dict"])

        # load optimizer state from checkpoint only when optimizer type is not changed.
        if (
            checkpoint["config"]["optimizer"]["type"]
            != self.config["optimizer"]["type"]
        ):
            self.logger.warning(
                "Warning: Optimizer type given in config file is different from that of checkpoint. "
                "Optimizer parameters not being resumed."
            )
        else:
            self.optimizer.load_state_dict(checkpoint["optimizer"])

        self.logger.info(
            "Checkpoint loaded. Resume training from epoch {}".format(self.start_epoch)
        )
